"""Module for handling the Amazon service"""

import base64
import hashlib
import json
import lzma
import os
import secrets
import struct
import time
import urllib
import uuid
from collections import defaultdict
from gettext import gettext as _
from typing import Any, Dict, Optional
from urllib.parse import parse_qs, urlencode, urlparse

import yaml

from lutris import settings
from lutris.exceptions import AuthenticationError, UnavailableGameError
from lutris.installer import AUTO_WIN32_EXE
from lutris.installer.installer_file import InstallerFile
from lutris.installer.installer_file_collection import InstallerFileCollection
from lutris.services.base import SERVICE_LOGIN, OnlineService
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util import system
from lutris.util.amazon.sds_proto2 import CompressionAlgorithm, HashAlgorithm, Manifest, ManifestHeader
from lutris.util.http import HTTPError, Request
from lutris.util.log import logger
from lutris.util.strings import slugify


class AmazonBanner(ServiceMedia):
    """Game logo"""

    service = "amazon"
    size = (200, 112)
    dest_path = os.path.join(settings.CACHE_DIR, "amazon/banners")
    file_patterns = ["%s.jpg"]
    api_field = "image"
    url_pattern = "%s"

    def get_media_url(self, details: Dict[str, Any]) -> Optional[str]:
        return details["product"]["productDetail"]["details"]["logoUrl"]


class AmazonGame(ServiceGame):
    """Representation of a Amazon game"""

    service = "amazon"

    @classmethod
    def new_from_amazon_game(cls, amazon_game):
        """Return a Amazon game instance from the API info"""
        service_game = AmazonGame()
        service_game.appid = str(amazon_game["id"])
        service_game.slug = slugify(amazon_game["product"]["title"])
        service_game.name = amazon_game["product"]["title"]
        service_game.details = json.dumps(amazon_game)
        return service_game


class AmazonService(OnlineService):
    """Service class for Amazon"""

    id = "amazon"
    name = _("Amazon")
    icon = "amazon"
    runner = "wine"
    has_extras = False
    drm_free = False
    medias = {"banner": AmazonBanner}
    default_format = "banner"

    login_window_width = 400
    login_window_height = 710

    marketplace_id = "ATVPDKIKX0DER"
    user_agent = "com.amazon.agslauncher.win/3.0.9202.1"

    amazon_api = "https://api.amazon.com"
    amazon_sds = "https://sds.amazon.com"
    amazon_gaming_graphql = "https://gaming.amazon.com/graphql"
    amazon_gaming_distribution = "https://gaming.amazon.com/api/distribution/v2/public"
    amazon_gaming_entitlements = "https://gaming.amazon.com/api/distribution/entitlements"

    client_id = None
    serial = None
    verifier = None

    redirect_uris = ["https://www.amazon.com/?"]

    cookies_path = os.path.join(settings.CACHE_DIR, ".amazon.auth")
    user_path = os.path.join(settings.CACHE_DIR, ".amazon.user")
    cache_path = os.path.join(settings.CACHE_DIR, "amazon-library.json")

    locale = "en-US"

    @property
    def credential_files(self):
        return [self.user_path, self.cookies_path]

    @property
    def login_url(self):
        """Return authentication URL"""
        self.verifier = self.generate_code_verifier()
        challenge = self.generate_challange(self.verifier)

        self.serial = self.generate_device_serial()
        self.client_id = self.generate_client_id(self.serial)

        arguments = {
            "openid.ns": "http://specs.openid.net/auth/2.0",
            "openid.claimed_id": "http://specs.openid.net/auth/2.0/identifier_select",
            "openid.identity": "http://specs.openid.net/auth/2.0/identifier_select",
            "openid.mode": "checkid_setup",
            "openid.oa2.scope": "device_auth_access",
            "openid.ns.oa2": "http://www.amazon.com/ap/ext/oauth/2",
            "openid.oa2.response_type": "code",
            "openid.oa2.code_challenge_method": "S256",
            "openid.oa2.client_id": f"device:{self.client_id}",
            "language": "en_US",
            "marketPlaceId": self.marketplace_id,
            "openid.return_to": "https://www.amazon.com",
            "openid.pape.max_auth_age": 0,
            "openid.assoc_handle": "amzn_sonic_games_launcher",
            "pageId": "amzn_sonic_games_launcher",
            "openid.oa2.code_challenge": challenge,
        }

        return "https://amazon.com/ap/signin?" + urlencode(arguments)

    def login_callback(self, url):
        """Get authentication token from Amazon"""
        if url.find("openid.oa2.authorization_code") > 0:
            logger.info("Got authorization code")

            # Parse auth code
            parsed = urlparse(url)
            query = parse_qs(parsed.query)
            auth_code = query["openid.oa2.authorization_code"][0]

            user_data = self.register_device(auth_code)
            user_data["token_obtain_time"] = time.time()

            self.save_user_data(user_data)

            SERVICE_LOGIN.fire(self)

    def is_connected(self):
        """Return whether the user is authenticated and if the service is available"""
        if not self.is_authenticated():
            return False

        return self.check_connection()

    def load(self):
        """Load the user game library from the Amazon API"""
        if not self.is_connected():
            logger.error("User not connected to Amazon")
            return
        games = [AmazonGame.new_from_amazon_game(game) for game in self.get_library()]
        for game in games:
            game.save()
        return games

    def save_user_data(self, user_data):
        """Save the user data file"""
        with open(self.user_path, "w", encoding="utf-8") as user_file:
            user_file.write(json.dumps(user_data))

    def load_user_data(self):
        """Load the user data file"""
        user_data = None

        if not os.path.exists(self.user_path):
            raise AuthenticationError(_("No Amazon user data available, please log in again"))

        with open(self.user_path, "r", encoding="utf-8") as user_file:
            user_data = json.load(user_file)

        return user_data

    def generate_code_verifier(self) -> bytes:
        code_verifier = secrets.token_bytes(32)
        code_verifier = base64.urlsafe_b64encode(code_verifier).rstrip(b"=")
        logger.info("Generated code_verifier: %s", code_verifier)
        return code_verifier

    def generate_challange(self, code_verifier: bytes) -> bytes:
        challenge_hash = hashlib.sha256(code_verifier)
        challenge = base64.urlsafe_b64encode(challenge_hash.digest()).rstrip(b"=")
        logger.info("Generated challange: %s", challenge)
        return challenge

    def generate_device_serial(self) -> str:
        serial = uuid.UUID(int=uuid.getnode()).hex.upper()
        logger.info("Generated serial: %s", serial)
        return serial

    def generate_client_id(self, serial) -> str:
        serialEx = f"{serial}#A2UMVHOX7UP4V7"
        clientId = serialEx.encode("ascii")
        clientIdHex = clientId.hex()
        logger.info("Generated client_id: %s", clientIdHex)
        return clientIdHex

    def register_device(self, code):
        """Register current device and return the user data"""
        logger.info("Registerring a device. ID: %s", self.client_id)
        data = {
            "auth_data": {
                "authorization_code": code,
                "client_domain": "DeviceLegacy",
                "client_id": self.client_id,
                "code_algorithm": "SHA-256",
                "code_verifier": self.verifier.decode("utf-8"),
                "use_global_authentication": False,
            },
            "registration_data": {
                "app_name": "AGSLauncher for Windows",
                "app_version": "1.0.0",
                "device_model": "Windows",
                "device_name": None,
                "device_serial": self.serial,
                "device_type": "A2UMVHOX7UP4V7",
                "domain": "Device",
                "os_version": "10.0.19044.0",
            },
            "requested_extensions": ["customer_info", "device_info"],
            "requested_token_type": ["bearer", "mac_dms"],
            "user_context_map": {},
        }

        url = f"{self.amazon_api}/auth/register"
        request = Request(url)

        try:
            request.post(json.dumps(data).encode())
        except HTTPError as ex:
            logger.exception("Failed http request %s: %s", url, ex)
            raise AuthenticationError(_("Unable to register device, please log in again")) from ex

        res_json = request.json
        logger.info("Successfully registered a device")
        user_data = res_json["response"]["success"]
        return user_data

    def is_token_expired(self):
        """Check if the stored token is expired"""
        user_data = self.load_user_data()

        token_obtain_time = user_data["token_obtain_time"]
        expires_in = user_data["tokens"]["bearer"]["expires_in"]

        if not token_obtain_time or not expires_in:
            raise AuthenticationError(_("Invalid token info found, please log in again"))

        return time.time() > token_obtain_time + int(expires_in)

    def refresh_token(self):
        """Refresh the token"""
        url = f"{self.amazon_api}/auth/token"
        logger.info("Refreshing token")

        user_data = self.load_user_data()

        headers = {
            "Accept": "application/json",
            "Accept-Language": "en_US",
            "User-Agent": self.user_agent,
            "Content-Type": "application/json",
            "charset": "utf-8",
        }

        refresh_token = user_data["tokens"]["bearer"]["refresh_token"]
        request_data = {
            "source_token": refresh_token,
            "source_token_type": "refresh_token",
            "requested_token_type": "access_token",
            "app_name": "AGSLauncher for Windows",
            "app_version": "1.0.0",
        }

        request = Request(url, headers=headers)

        try:
            request.post(json.dumps(request_data).encode())
        except HTTPError as ex:
            logger.exception("Failed http request %s: %s", url, ex)
            raise AuthenticationError(_("Unable to refresh token, please log in again")) from ex

        res_json = request.json

        user_data["tokens"]["bearer"]["access_token"] = res_json["access_token"]
        user_data["tokens"]["bearer"]["expires_in"] = res_json["expires_in"]
        user_data["token_obtain_time"] = time.time()

        self.save_user_data(user_data)

    def get_access_token(self):
        """Return the access token and refresh the session if required"""
        if self.is_token_expired():
            self.refresh_token()

        user_data = self.load_user_data()
        access_token = user_data["tokens"]["bearer"]["access_token"]

        return access_token

    def check_connection(self):
        """Check if the connection with Amazon is available"""

        try:
            access_token = self.get_access_token()
        except Exception:
            return False

        headers = {
            "Accept": "application/json",
            "Accept-Language": "en_US",
            "User-Agent": self.user_agent,
            "Authorization": f"bearer {access_token}",
        }

        url = f"{self.amazon_api}/user/profile"
        request = Request(url, headers=headers)

        try:
            request.get()
        except HTTPError as ex:
            # Do not raise exception here, should be managed from the caller
            logger.exception("Failed http request %s: %s", url, ex)
            return False

        return True

    def get_library(self):
        """Return the user's library of Amazon games"""
        if system.path_exists(self.cache_path):
            logger.debug("Returning cached Amazon library")
            with open(self.cache_path, "r", encoding="utf-8") as amazon_cache:
                return json.load(amazon_cache)

        access_token = self.get_access_token()

        user_data = self.load_user_data()
        serial = user_data["extensions"]["device_info"]["device_serial_number"]

        games_by_asin = defaultdict(list)
        next_token = None
        while True:
            request_data = self.get_sync_request_data(serial, next_token)

            json_data = self.request_entitlements(
                "com.amazon.animusdistributionservice.entitlement.AnimusEntitlementsService.GetEntitlements",
                access_token,
                request_data,
            )

            for game_json in json_data["entitlements"]:
                product = game_json["product"]

                if "id" not in product:
                    logger.error("Amazon game encountered with no ID; skipping: %s", game_json)
                    continue

                # Some games have no ASIN; we'll have to skip dedupping these by ASIN.
                asin = product.get("asin") or ""
                games_by_asin[asin].append(game_json)

            if "nextToken" not in json_data:
                break

            logger.info("Got next token in response, making next request")
            next_token = json_data["nextToken"]

        # We need to use the ID for compatibility with earlier Lutris releases, but also
        # some games do not have ASINs. We still deduplicate by ASIN when we can, since the same
        # game can have two different IDs.

        asinless = games_by_asin.pop("", [])
        dedupped = [sorted(gl, key=lambda g: g["id"])[0] for gl in games_by_asin.values()]
        games = asinless + dedupped

        with open(self.cache_path, "w", encoding="utf-8") as amazon_cache:
            json.dump(games, amazon_cache)

        return games

    def request_distribution(self, target, token, body):
        headers = {
            "X-Amz-Target": target,
            "x-amzn-token": token,
            "UserAgent": "com.amazon.agslauncher.win/3.0.9202.1",
            "Content-Type": "application/json",
            "Content-Encoding": "amz-1.0",
        }
        request = Request(self.amazon_gaming_distribution, headers=headers)
        request.post(json.dumps(body).encode())
        return request.json

    def get_sync_request_data(self, serial, next_token=None, sync_point=None):
        request_data = {
            "Operation": "GetEntitlements",
            "clientId": "Sonic",
            "syncPoint": sync_point,
            "nextToken": next_token,
            "maxResults": 50,
            "productIdFilter": None,
            "keyId": "d5dc8b8b-86c8-4fc4-ae93-18c0def5314d",
            "hardwareHash": hashlib.sha256(serial.encode()).hexdigest().upper(),
        }

        return request_data

    def request_sds(self, target, token, body):
        headers = {
            "X-Amz-Target": target,
            "x-amzn-token": token,
            "User-Agent": self.user_agent,
            "Content-Type": "application/json",
            "Content-Encoding": "amz-1.0",
        }

        url = f"{self.amazon_sds}/amazon/"
        request = Request(url, headers=headers)
        request.post(json.dumps(body).encode())
        return request.json

    def request_entitlements(self, target, token, body):
        headers = {
            "X-Amz-Target": target,
            "x-amzn-token": token,
            "User-Agent": self.user_agent,
            "Content-Type": "application/json",
            "Content-Encoding": "amz-1.0",
        }

        url = f"{self.amazon_gaming_entitlements}"
        request = Request(url, headers=headers)
        request.post(json.dumps(body).encode())
        return request.json

    def get_game_manifest_info(self, game_id):
        """Get a game manifest information"""
        access_token = self.get_access_token()

        request_data = {
            "entitlementId": game_id,
            "Operation": "GetGameDownload",
        }

        try:
            response = self.request_distribution(
                "com.amazon.animusdistributionservice.external.AnimusDistributionService.GetGameDownload",
                access_token,
                request_data,
            )
        except HTTPError as ex:
            # Do not raise exception here, should be managed from the caller
            logger.exception("There was an error getting game '%s' manifest: %s", game_id, ex)
            raise UnavailableGameError(_("Unable to get game manifest info")) from ex

        return response

    def get_game_manifest(self, manifest_info):
        """Get a game manifest"""
        headers = {
            "User-Agent": self.user_agent,
        }

        url = manifest_info["downloadUrl"]
        url = urllib.parse.urlparse(url)
        url = url._replace(path=url.path + "/manifest.proto")
        url = urllib.parse.urlunparse(url)

        request = Request(url, headers=headers)

        try:
            request.get()
        except HTTPError as ex:
            logger.exception("Failed http request %s: %s", url, ex)
            raise UnavailableGameError(_("Unable to get game manifest")) from ex

        content = request.content

        header_size = struct.unpack(">I", content[:4])[0]

        header = ManifestHeader()
        header.decode(content[4 : 4 + header_size])

        if header.compression.algorithm == CompressionAlgorithm.none:
            raw_manifest = content[4 + header_size :]
        elif header.compression.algorithm == CompressionAlgorithm.lzma:
            raw_manifest = lzma.decompress(content[4 + header_size :])
        else:
            logger.error("Unknown compression algorithm found in manifest")
            raise UnavailableGameError(_("Unknown compression algorithm found in manifest"))

        manifest = Manifest()
        manifest.decode(raw_manifest)

        return manifest

    def get_file_patch(self, access_token, game_id, version, file_hashes):
        request_data = {
            "Operation": "GetPatches",
            "versionId": version,
            "fileHashes": file_hashes,
            "deltaEncodings": ["FUEL_PATCH", "NONE"],
            "adgGoodId": game_id,
        }

        try:
            return self.request_sds(
                "com.amazonaws.gearbox.softwaredistribution.service.model.SoftwareDistributionService.GetPatches",
                access_token,
                request_data,
            )
        except HTTPError as ex:
            # Do not raise exception here, should be managed from the caller
            logger.exception("There was an error getting '%s' patches: %s", game_id, ex)
            raise UnavailableGameError(_("Unable to get the patches of game '%s'") % game_id) from ex

    def get_game_patches(self, game_id, version, file_list):
        """Get game files"""
        access_token = self.get_access_token()

        def get_batches(to_batch, batch_size):
            i = 0
            while i < len(to_batch):
                yield to_batch[i : i + batch_size]
                i += batch_size

        batches = get_batches(file_list, 500)
        patches = []

        for batch in batches:
            response = self.get_file_patch(access_token, game_id, version, batch)
            patches += response["patches"]

        return patches

    def structure_manifest_data(self, manifest):
        """Transform the manifest to more convenient data structures"""
        file_dict = {}
        directories = []
        hashpairs = []
        file_paths = set()

        for __, package in enumerate(manifest.packages):
            for __, file in enumerate(package.files):
                file_hash = file.hash.value.hex()
                file_path = file.path.decode().replace("\\", "/")
                file_paths.add(file_path)

                if file_hash in file_dict:
                    file_dict[file_hash]["paths"].append(file_path)
                else:
                    file_dict[file_hash] = {
                        "paths": [file_path],
                        "size": file.size,
                        "url": None,
                    }

                hashpairs.append(
                    {
                        "sourceHash": None,
                        "targetHash": {"value": file_hash, "algorithm": HashAlgorithm.get_name(file.hash.algorithm)},
                    }
                )
            for __, directory in enumerate(package.dirs):
                if directory.path is not None:
                    dir_path = directory.path.decode().replace("\\", "/")
                    if dir_path not in file_paths:
                        directories.append(dir_path)

        return file_dict, directories, hashpairs

    def get_game_files(self, game_id):
        """Get the game file list"""

        manifest_info = self.get_game_manifest_info(game_id)
        manifest = self.get_game_manifest(manifest_info)

        file_dict, directories, hashpairs = self.structure_manifest_data(manifest)

        for file_hash, file in file_dict.items():
            url = manifest_info["downloadUrl"]
            url = urllib.parse.urlparse(url)
            url = url._replace(path=url.path + "/files/" + file_hash)
            url = urllib.parse.urlunparse(url)
            file["url"] = url

        return file_dict, directories

    def get_exe_and_game_args(self, fuel_url):
        """Get and parse the fuel.json file"""
        headers = {
            "User-Agent": self.user_agent,
        }

        request = Request(fuel_url, headers=headers)

        try:
            request.get()
        except HTTPError as ex:
            logger.error("Failed http request %s: %s", fuel_url, ex)
            raise UnavailableGameError(_("Unable to get fuel.json file.")) from ex

        try:
            res_yaml_text = request.text
            res_json = yaml.safe_load(res_yaml_text)
        except Exception as ex:
            # Maybe it can be parsed as plain JSON. May as well try it.
            try:
                logger.exception("Unparseable yaml response from %s: %s\n%s", fuel_url, ex, res_yaml_text)
                res_json = json.loads(res_yaml_text)
            except Exception:
                raise UnavailableGameError(_("Invalid response from Amazon APIs")) from ex

        if res_json["Main"] is None or res_json["Main"]["Command"] is None:
            return None, None

        game_cmd = res_json["Main"]["Command"].replace("\\", "/")
        game_args = ""

        if "Args" in res_json["Main"] and res_json["Main"]["Args"]:
            for arg in res_json["Main"]["Args"]:
                game_args += arg if game_args == "" else " " + arg

        return game_cmd, game_args

    def get_game_cmd_line(self, fuel_url):
        """Get the executable path and the arguments for run the game"""
        game_cmd = None
        game_args = None

        if fuel_url is not None:
            game_cmd, game_args = self.get_exe_and_game_args(fuel_url)

        if game_cmd is None:
            game_cmd = AUTO_WIN32_EXE

        if game_args is None:
            game_args = ""

        return game_cmd, game_args

    def get_installer_files(self, installer, _installer_file_id, _selected_extras):
        try:
            file_dict, __ = self.get_game_files(installer.service_appid)
        except HTTPError as err:
            raise UnavailableGameError(_("Couldn't load the downloads for this game")) from err

        files = []
        for file_hash, file in file_dict.items():
            file_name = os.path.basename(file["paths"][0])
            files.append(
                InstallerFile(
                    installer.game_slug, file_hash, {"url": file["url"], "filename": file_name, "size": file["size"]}
                )
            )
        # return should be a list of files, so we return a list containing a InstallerFileCollection
        file_collection = InstallerFileCollection(installer.game_slug, "amazongame", files)
        return [file_collection], []

    def get_installed_slug(self, db_game):
        details = json.loads(db_game["details"])
        return slugify(details["product"]["title"])

    def generate_installer(self, db_game):
        """Generate a installer for the Amazon game"""
        details = json.loads(db_game["details"])
        game_id = details["id"]

        manifest_info = self.get_game_manifest_info(game_id)
        manifest = self.get_game_manifest(manifest_info)

        file_dict, directories, hashpairs = self.structure_manifest_data(manifest)

        installer = [
            {"task": {"name": "create_prefix"}},
            {"mkdir": "$GAMEDIR/drive_c/game"},
            {"autosetup_amazon": {"files": file_dict, "directories": directories}},
        ]

        # try to get fuel file that contain the main exe
        fuel_file = [k for k, v in file_dict.items() if any("fuel.json" in p for p in v["paths"])]
        fuel_url = None
        if fuel_file:
            fuel_url = manifest_info["downloadUrl"]
            fuel_url = urllib.parse.urlparse(fuel_url)
            fuel_url = fuel_url._replace(path=fuel_url.path + "/files/" + list(fuel_file)[0])
            fuel_url = urllib.parse.urlunparse(fuel_url)

        game_cmd, game_args = self.get_game_cmd_line(fuel_url)
        logger.info("game cmd line: %s %s", game_cmd, game_args)

        return {
            "name": details["product"]["title"],
            "version": _("Amazon Prime Gaming"),
            "slug": slugify(details["product"]["title"]),
            "game_slug": slugify(details["product"]["title"]),
            "runner": self.get_installed_runner_name(db_game),
            "script": {
                "game": {
                    "exe": f"$GAMEDIR/drive_c/game/{game_cmd}",
                    "args": game_args,
                    "prefix": "$GAMEDIR",
                    "working_dir": "$GAMEDIR/drive_c/game",
                },
                "system": {},
                "files": [{"amazongame": "N/A:Select the installer from Amazon Games"}],
                "installer": installer,
            },
        }

    def get_installed_runner_name(self, db_game):
        return self.runner



import os
import subprocess
from gettext import gettext as _
from typing import Any, Dict, List, Optional

from lutris import settings
from lutris.config import LutrisConfig
from lutris.exceptions import MissingExecutableError
from lutris.runners.runner import Runner
from lutris.util import system
from lutris.util.strings import split_arguments

_supported_scale_factors = {
    "hq": ["2", "3"],
    "edge": ["2", "3"],
    "advmame": ["2", "3"],
    "sai": ["2"],
    "supersai": ["2"],
    "supereagle": ["2"],
    "dotmatrix": ["2"],
    "tv2x": ["2"],
}


def _get_opengl_warning(_option_key: str, config: LutrisConfig) -> Optional[str]:
    runner_config = config.runner_config
    if "scaler" in runner_config and "renderer" in runner_config:
        renderer = runner_config["renderer"]
        if renderer and renderer != "software":
            scaler = runner_config["scaler"]
            if scaler and scaler != "normal":
                return _("<b>Warning</b> Scalers may not work with OpenGL rendering.")

    return None


def _get_scale_factor_warning(_option_key: str, config: LutrisConfig) -> Optional[str]:
    """Generate a warning message for when the scaler and scale-factor can't be used together."""
    runner_config = config.runner_config
    if "scaler" in runner_config and "scale-factor" in runner_config:
        scaler = runner_config["scaler"]
        if scaler in _supported_scale_factors:
            scale_factor = runner_config["scale-factor"]
            if scale_factor not in _supported_scale_factors[scaler]:
                return _("<b>Warning</b> The '%s' scaler does not work with a scale factor of %s.") % (
                    scaler,
                    scale_factor,
                )

    return None


class scummvm(Runner):
    description = _("Engine for point-and-click games.")
    human_name = _("ScummVM")
    platforms = [_("Linux")]
    runnable_alone = True
    runner_executable = "scummvm/bin/scummvm"
    flatpak_id = "org.scummvm.ScummVM"
    game_options = [
        {"option": "game_id", "type": "string", "label": _("Game identifier")},
        {"option": "path", "type": "directory", "label": _("Game files location")},
        {
            "option": "args",
            "type": "string",
            "label": _("Arguments"),
            "help": _("Command line arguments used when launching the game"),
        },
    ]

    option_map = {
        "aspect": "--aspect-ratio",
        "subtitles": "--subtitles",
        "fullscreen": "--fullscreen",
        "scaler": "--scaler=%s",
        "scale-factor": "--scale-factor=%s",
        "renderer": "--renderer=%s",
        "render-mode": "--render-mode=%s",
        "stretch-mode": "--stretch-mode=%s",
        "filtering": "--filtering",
        "platform": "--platform=%s",
        "engine-speed": "--engine-speed=%s",
        "talk-speed": "--talkspeed=%s",
        "dimuse-tempo": "--dimuse-tempo=%s",
        "music-tempo": "--tempo=%s",
        "opl-driver": "--opl-driver=%s",
        "output-rate": "--output-rate=%s",
        "music-driver": "--music-driver=%s",
        "multi-midi": "--multi-midi",
        "midi-gain": "--midi-gain=%s",
        "soundfont": "--soundfont=%s",
        "music-volume": "--music-volume=%s",
        "sfx-volume": "--sfx-volume=%s",
        "speech-volume": "--speech-volume=%s",
        "native-mt32": "--native-mt32",
        "enable-gs": "--enable-gs",
        "joystick": "--joystick=%s",
        "language": "--language=%s",
        "alt-intro": "--alt-intro",
        "copy-protection": "--copy-protection",
        "demo-mode": "--demo-mode",
        "debug-level": "--debug-level=%s",
        "debug-flags": "--debug-flags=%s",
    }

    option_empty_map = {"fullscreen": "--no-fullscreen"}

    runner_options = [
        {
            "option": "fullscreen",
            "section": _("Graphics"),
            "label": _("Fullscreen"),
            "type": "bool",
            "default": True,
        },
        {
            "option": "subtitles",
            "section": _("Graphics"),
            "label": _("Enable subtitles"),
            "type": "bool",
            "default": False,
            "help": ("Enable subtitles for games with voice"),
        },
        {
            "option": "aspect",
            "section": _("Graphics"),
            "label": _("Aspect ratio correction"),
            "type": "bool",
            "default": True,
            "help": _(
                "Most games supported by ScummVM were made for VGA "
                "display modes using rectangular pixels. Activating "
                "this option for these games will preserve the 4:3 "
                "aspect ratio they were made for."
            ),
        },
        {
            "option": "scaler",
            "section": _("Graphics"),
            "label": _("Graphic scaler"),
            "type": "choice",
            "default": "normal",
            "choices": [
                ("normal", "normal"),
                ("hq", "hq"),
                ("edge", "edge"),
                ("advmame", "advmame"),
                ("sai", "sai"),
                ("supersai", "supersai"),
                ("supereagle", "supereagle"),
                ("pm", "pm"),
                ("dotmatrix", "dotmatrix"),
                ("tv2x", "tv2x"),
            ],
            "warning": _get_opengl_warning,
            "help": _(
                "The algorithm used to scale up the game's base resolution, resulting in different visual styles. "
            ),
        },
        {
            "option": "scale-factor",
            "section": _("Graphics"),
            "label": _("Scale factor"),
            "type": "choice",
            "default": "3",
            "choices": [
                ("1", "1"),
                ("2", "2"),
                ("3", "3"),
                ("4", "4"),
                ("5", "5"),
            ],
            "help": _(
                "Changes the resolution of the game. "
                "For example, a 2x scale will take a 320x200 "
                "resolution game and scale it up to 640x400. "
            ),
            "warning": _get_scale_factor_warning,
        },
        {
            "option": "renderer",
            "section": _("Graphics"),
            "label": _("Renderer"),
            "type": "choice",
            "choices": [
                (_("Auto"), ""),
                (_("Software"), "software"),
                (_("OpenGL"), "opengl"),
                (_("OpenGL (with shaders)"), "opengl_shaders"),
            ],
            "default": "",
            "advanced": True,
            "help": _("Changes the rendering method used for 3D games."),
        },
        {
            "option": "render-mode",
            "section": _("Graphics"),
            "label": _("Render mode"),
            "type": "choice",
            "choices": [
                (_("Auto"), ""),
                (_("Hercules (Green)"), "hercGreen"),
                (_("Hercules (Amber)"), "hercAmber"),
                (_("CGA"), "cga"),
                (_("EGA"), "ega"),
                (_("VGA"), "vga"),
                (_("Amiga"), "amiga"),
                (_("FM Towns"), "fmtowns"),
                (_("PC-9821"), "pc9821"),
                (_("PC-9801"), "pc9801"),
                (_("Apple IIgs"), "2gs"),
                (_("Atari ST"), "atari"),
                (_("Macintosh"), "macintosh"),
            ],
            "default": "",
            "advanced": True,
            "help": _("Changes the graphics hardware the game will target, if the game supports this."),
        },
        {
            "option": "stretch-mode",
            "section": _("Graphics"),
            "label": _("Stretch mode"),
            "type": "choice",
            "choices": [
                (_("Auto"), ""),
                (_("Center"), "center"),
                (_("Pixel Perfect"), "pixel-perfect"),
                (_("Even Pixels"), "even-pixels"),
                (_("Stretch"), "stretch"),
                (_("Fit"), "fit"),
                (_("Fit (force aspect ratio)"), "fit_force_aspect"),
            ],
            "default": "",
            "advanced": True,
            "help": _("Changes how the game is placed when the window is resized."),
        },
        {
            "option": "filtering",
            "section": _("Graphics"),
            "label": _("Filtering"),
            "type": "bool",
            "help": _(
                "Uses bilinear interpolation instead of nearest neighbor "
                "resampling for the aspect ratio correction and stretch mode."
            ),
            "default": False,
            "advanced": True,
        },
        {
            "option": "datadir",
            "label": _("Data directory"),
            "type": "directory",
            "help": _("Defaults to share/scummvm if unspecified."),
            "advanced": True,
        },
        {
            "option": "platform",
            "type": "string",
            "label": _("Platform"),
            "help": _(
                "Specifes platform of game. Allowed values: 2gs, 3do, acorn, amiga, atari, c64, "
                "fmtowns, nes, mac, pc pc98, pce, segacd, wii, windows"
            ),
            "advanced": True,
        },
        {
            "option": "joystick",
            "type": "string",
            "label": _("Joystick"),
            "help": _("Enables joystick input (default: 0 = first joystick)"),
            "advanced": True,
        },
        {
            "option": "language",
            "type": "string",
            "label": _("Language"),
            "help": _("Selects language (en, de, fr, it, pt, es, jp, zh, kr, se, gb, hb, ru, cz)"),
            "advanced": True,
        },
        {
            "option": "engine-speed",
            "type": "string",
            "label": _("Engine speed"),
            "help": _(
                "Sets frames per second limit (0 - 100) for Grim Fandango or Escape from Monkey Island (default: 60)."
            ),
            "advanced": True,
        },
        {
            "option": "talk-speed",
            "type": "string",
            "label": _("Talk speed"),
            "help": _("Sets talk speed for games (default: 60)"),
            "advanced": True,
        },
        {
            "option": "music-tempo",
            "type": "string",
            "section": _("Audio"),
            "label": _("Music tempo"),
            "help": _("Sets music tempo (in percent, 50-200) for SCUMM games (default: 100)"),
            "advanced": True,
        },
        {
            "option": "dimuse-tempo",
            "type": "string",
            "section": _("Audio"),
            "label": _("Digital iMuse tempo"),
            "help": _("Sets internal Digital iMuse tempo (10 - 100) per second (default: 10)"),
            "advanced": True,
        },
        {
            "option": "music-driver",
            "section": _("Audio"),
            "label": _("Music driver"),
            "type": "choice",
            "choices": [
                ("null", "null"),
                ("auto", "auto"),
                ("seq", "seq"),
                ("sndio", "sndio"),
                ("alsa", "alsa"),
                ("fluidsynth", "fluidsynth"),
                ("mt32", "mt32"),
                ("adlib", "adlib"),
                ("pcspk", "pcspk"),
                ("pcjr", "pcjr"),
                ("cms", "cms"),
                ("timidity", "timidity"),
            ],
            "help": _("Specifies the device ScummVM uses to output audio."),
            "advanced": True,
        },
        {
            "option": "output-rate",
            "section": _("Audio"),
            "label": _("Output rate"),
            "type": "choice",
            "choices": [
                ("11025", "11025"),
                ("22050", "22050"),
                ("44100", "44100"),
            ],
            "help": _("Selects output sample rate in Hz."),
            "advanced": True,
        },
        {
            "option": "opl-driver",
            "section": _("Audio"),
            "label": _("OPL driver"),
            "type": "choice",
            "choices": [
                ("auto", "auto"),
                ("mame", "mame"),
                ("db", "db"),
                ("nuked", "nuked"),
                ("alsa", "alsa"),
                ("op2lpt", "op2lpt"),
                ("op3lpt", "op3lpt"),
                ("rwopl3", "rwopl3"),
            ],
            "help": _(
                "Chooses which emulator is used by ScummVM when the AdLib emulator is chosen as the Preferred device."
            ),
            "advanced": True,
        },
        {
            "option": "music-volume",
            "type": "string",
            "section": _("Audio"),
            "label": _("Music volume"),
            "help": _("Sets the music volume, 0-255 (default: 192)"),
            "advanced": True,
        },
        {
            "option": "sfx-volume",
            "type": "string",
            "section": _("Audio"),
            "label": _("SFX volume"),
            "help": _("Sets the sfx volume, 0-255 (default: 192)"),
            "advanced": True,
        },
        {
            "option": "speech-volume",
            "type": "string",
            "section": _("Audio"),
            "label": _("Speech volume"),
            "help": _("Sets the speech volume, 0-255 (default: 192)"),
            "advanced": True,
        },
        {
            "option": "midi-gain",
            "type": "string",
            "section": _("Audio"),
            "label": _("MIDI gain"),
            "help": _("Sets the gain for MIDI playback. 0-1000 (default: 100)"),
            "advanced": True,
        },
        {
            "option": "soundfont",
            "section": _("Audio"),
            "type": "string",
            "label": _("Soundfont"),
            "help": _("Specifies the path to a soundfont file."),
            "advanced": True,
        },
        {
            "option": "multi-midi",
            "section": _("Audio"),
            "label": _("Mixed AdLib/MIDI mode"),
            "type": "bool",
            "default": False,
            "help": _("Combines MIDI music with AdLib sound effects."),
            "advanced": True,
        },
        {
            "option": "native-mt32",
            "section": _("Audio"),
            "label": _("True Roland MT-32"),
            "type": "bool",
            "default": False,
            "help": _(
                "Tells ScummVM that the MIDI device is an actual Roland MT-32, "
                "LAPC-I, CM-64, CM-32L, CM-500 or other MT-32 device."
            ),
            "advanced": True,
        },
        {
            "option": "enable-gs",
            "section": _("Audio"),
            "label": _("Enable Roland GS"),
            "type": "bool",
            "default": False,
            "help": _(
                "Tells ScummVM that the MIDI device is a GS device that has "
                "an MT-32 map, such as an SC-55, SC-88 or SC-8820."
            ),
            "advanced": True,
        },
        {
            "option": "alt-intro",
            "type": "bool",
            "label": _("Use alternate intro"),
            "help": _("Uses alternative intro for CD versions"),
            "advanced": True,
        },
        {
            "option": "copy-protection",
            "type": "bool",
            "label": _("Copy protection"),
            "help": _("Enables copy protection"),
            "advanced": True,
        },
        {
            "option": "demo-mode",
            "type": "bool",
            "label": _("Demo mode"),
            "help": _("Starts demo mode of Maniac Mansion or The 7th Guest"),
            "advanced": True,
        },
        {
            "option": "debug-level",
            "type": "string",
            "section": _("Debugging"),
            "label": _("Debug level"),
            "help": _("Sets debug verbosity level"),
            "advanced": True,
        },
        {
            "option": "debug-flags",
            "type": "string",
            "section": _("Debugging"),
            "label": _("Debug flags"),
            "help": _("Enables engine specific debug flags"),
            "advanced": True,
        },
    ]

    @property
    def game_path(self):
        return self.game_config.get("path")

    def get_extra_libs(self) -> List[str]:
        """Scummvm runner ships additional libraries, they may be removed in a future version."""
        try:
            base_runner_path = os.path.join(settings.RUNNER_DIR, "scummvm")
            if self.get_executable().startswith(base_runner_path):
                path = os.path.join(settings.RUNNER_DIR, "scummvm/lib")
                if system.path_exists(path):
                    return [path]
        except MissingExecutableError:
            pass

        return []

    def get_command(self) -> List[str]:
        command = super().get_command()
        if not command:
            return []
        if "flatpak" in command[0]:
            return command

        data_dir = self.get_scummvm_data_dir()

        return command + [
            "--extrapath=%s" % data_dir,
            "--themepath=%s" % data_dir,
        ]

    def get_scummvm_data_dir(self) -> str:
        data_dir = self.runner_config.get("datadir")

        if data_dir is None:
            root_dir = os.path.dirname(os.path.dirname(self.get_executable()))
            data_dir = os.path.join(root_dir, "share/scummvm")

        return data_dir

    def get_run_data(self) -> Dict[str, Any]:
        env = self.get_env()
        lib_paths = filter(None, self.get_extra_libs() + [env.get("LD_LIBRARY_PATH")])
        if lib_paths:
            env["LD_LIBRARY_PATH"] = os.pathsep.join(lib_paths)

        return {"env": env, "command": self.get_command()}

    def inject_runner_option(self, command, key, cmdline, cmdline_empty=None):
        value = self.runner_config.get(key)
        if value:
            if "%s" in cmdline:
                command.append(cmdline % value)
            else:
                command.append(cmdline)
        elif cmdline_empty:
            command.append(cmdline_empty)

    def play(self):
        command = self.get_command()
        for option, cmdline in self.option_map.items():
            self.inject_runner_option(command, option, cmdline, self.option_empty_map.get(option))
        command.append("--path=%s" % self.game_path)
        args = self.game_config.get("args") or ""
        for arg in split_arguments(args):
            command.append(arg)
        if self.game_config.get("game_id"):
            command.append(self.game_config.get("game_id"))
        output = {"command": command}

        extra_libs = self.get_extra_libs()
        if extra_libs:
            output["ld_library_path"] = os.pathsep.join(extra_libs)

        return output

    def get_game_list(self) -> List[List[str]]:
        """Return the entire list of games supported by ScummVM."""
        with subprocess.Popen(
            self.get_command() + ["--list-games"], stdout=subprocess.PIPE, encoding="utf-8", universal_newlines=True
        ) as scummvm_process:
            scumm_output = scummvm_process.communicate()[0]
            game_list = str.split(scumm_output, "\n")
        game_array = []
        game_list_start = False
        for game in game_list:
            if game_list_start:
                if len(game) > 1:
                    dir_limit = game.index(" ")
                else:
                    dir_limit = None
                if dir_limit is not None:
                    game_dir = game[0:dir_limit]
                    game_name = game[dir_limit + 1 : len(game)].strip()
                    game_array.append([game_dir, game_name])
            # The actual list is below a separator
            if game.startswith("-----"):
                game_list_start = True
        return game_array



import json
import os
import random
import time
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Tuple

from lutris.database.services import ServiceGameCollection
from lutris.util import system
from lutris.util.http import HTTPError, download_file
from lutris.util.log import logger
from lutris.util.portals import TrashPortal

if TYPE_CHECKING:
    from lutris.services.base import BaseService


class MediaPath:
    """An object to describe a media file along with the media size- which
    is defined by Lutris, not the size of the image in the file. Note that
    the file may not exist."""

    def __init__(self, path: str, service_media: "ServiceMedia", size: Optional[Tuple[int, int]] = None):
        self.path = path
        self.service_media = service_media
        self.size = size or service_media.size

    @property
    def width(self) -> int:
        return self.size[0]

    @property
    def height(self) -> int:
        return self.size[1]

    @property
    def exists(self) -> bool:
        return system.path_exists(self.path, exclude_empty=True) and os.path.isfile(self.path)

    def scale_to_fit(self, max_size: Tuple[int, int]) -> "MediaPath":
        if max_size != self.size:
            x_factor = max_size[0] / self.width
            y_factor = max_size[1] / self.height
            factor = min(x_factor, y_factor)
            scaled_width = int(self.width * factor)
            scaled_height = int(self.height * factor)
            return MediaPath(self.path, self.service_media, (scaled_width, scaled_height))
        return self

    def __repr__(self) -> str:
        return self.path


def resolve_media_path(possible_paths: List[MediaPath]) -> MediaPath:
    """Selects the best path from a list of paths to media. This will take the first
    one that exists and has contents, or the just first one if none are usable."""
    if len(possible_paths) > 1:
        for mp in possible_paths:
            if mp.exists:
                return mp
    elif not possible_paths:
        raise ValueError("resolve_media_path() requires at least one path.")

    return possible_paths[0]


class ServiceMedia:
    """Information about the service's media format"""

    service = NotImplemented
    size = NotImplemented
    source = "remote"  # set to local if the files don't need to be downloaded
    visible = True  # This media should be displayed as an option in the UI
    dest_path = NotImplemented
    file_patterns = NotImplemented
    api_field = NotImplemented
    url_pattern = "%s"
    can_be_fallback = True

    def __init__(self):
        if self.dest_path and not system.path_exists(self.dest_path):
            os.makedirs(self.dest_path)

    def get_filename(self, slug):
        return self.file_patterns[0] % slug

    def get_possible_media_paths(self, slug: str) -> List[MediaPath]:
        """Returns a list of each path where the media might be found. At most one of these should
        be found, but they are in a priority order - the first is in the preferred format."""
        return [MediaPath(os.path.join(self.dest_path, pattern % slug), self) for pattern in self.file_patterns]

    def get_fallback_media_path(
        self, services: Iterable[Tuple["BaseService", Callable[[], str]]]
    ) -> Optional[MediaPath]:
        """Returns the media path to use when none of the possible paths (above) actually exist.
        This may be scaled down so it no taller than this media's height.

        This method finds the media that is nearest to the size of this media, and
        then evaluates the provided callables to obtain slugs, starting with the nearest
        match. As soon as it finds a media that exists, it will return it. If it finds
        none, it returns None.
        """

        medias = [(mt(), t[1]) for t in services for mt in t[0].medias.values()]

        def similarity(media: Tuple[ServiceMedia, Callable[[], str]]) -> int:
            diff = abs(media[0].size[1] - self.size[1])
            return diff if media[0].size[1] >= self.size[1] else diff + 1000

        for media, slug_function in sorted(medias, key=similarity):
            if media.can_be_fallback:
                slug = slug_function()
                if slug:
                    for mp in media.get_possible_media_paths(slug):
                        if mp.exists:
                            return mp.scale_to_fit(self.size)

        return None

    def trash_media(
        self,
        slug: str,
        completion_function: Optional[TrashPortal.CompletionFunction] = None,
        error_function: Optional[TrashPortal.ErrorFunction] = None,
    ) -> None:
        """Sends each media file for a game to the trash, and invokes callsbacks when this
        has been completed or has failed."""
        paths = [mp.path for mp in self.get_possible_media_paths(slug) if os.path.exists(mp.path)]
        if paths:
            TrashPortal(paths, completion_function=completion_function, error_function=error_function)
        elif completion_function:
            completion_function()

    def get_media_url(self, details: Dict[str, Any]) -> Optional[str]:
        if self.api_field not in details:
            logger.warning("No field '%s' in API game %s", self.api_field, details)
            return None
        if not details[self.api_field]:
            return None
        return self.url_pattern % details[self.api_field]

    def get_media_urls(self) -> Dict[str, str]:
        """Return URLs for icons and logos from a service"""
        if self.source == "local":
            return {}
        service_games = ServiceGameCollection.get_for_service(self.service)
        medias: Dict[str, str] = {}
        for game in service_games:
            if not game["details"]:
                continue
            details = json.loads(game["details"])
            media_url = self.get_media_url(details)
            if not media_url:
                continue
            medias[game["slug"]] = media_url
        return medias

    def download(self, slug, url):
        """Downloads the banner if not present"""
        if not url:
            return
        cache_path = os.path.join(self.dest_path, self.get_filename(slug))
        if system.path_exists(cache_path, exclude_empty=True):
            return
        if system.path_exists(cache_path):
            cache_stats = os.stat(cache_path)
            # Empty files have a life time between 1 and 2 weeks, retry them after
            if time.time() - cache_stats.st_mtime < 3600 * 24 * random.choice(range(7, 15)):
                return cache_path
            os.unlink(cache_path)
        try:
            return download_file(url, cache_path, raise_errors=True)
        except HTTPError as ex:
            logger.error("Failed to download %s: %s", url, ex)

    @property
    def custom_media_storage_size(self):
        """The size this media is stored in when customized; we accept
        whatever we get when we download the media, however."""
        return self.size

    @property
    def config_ui_size(self):
        """The size this media should be shown at when in the configuration UI."""
        return self.size

    def run_system_update_desktop_icons(self):
        """Update the desktop, if this media type appears there. Most don't."""

    def render(self):
        """Used if the media requires extra processing"""



"""Misc widgets used in the GUI."""

# Standard Library
import os
import shlex
import urllib.parse
from gettext import gettext as _

# Third Party Libraries
from gi.repository import GLib, GObject, Gtk, Pango

from lutris.gui.widgets.utils import open_uri

# Lutris Modules
from lutris.util import system
from lutris.util.linux import LINUX_SYSTEM

# MyPy does not like GTK's notion of multiple inheritancs, but
# we don't control this, so we'll suppress type checking.


class SlugEntry(Gtk.Entry, Gtk.Editable):  # type:ignore[misc]
    def do_insert_text(self, new_text, length, position):
        """Filter inserted characters to only accept alphanumeric and dashes"""
        new_text = "".join([c for c in new_text if c.isalnum() or c == "-"]).lower()
        length = len(new_text)
        self.get_buffer().insert_text(position, new_text, length)
        return position + length


class NumberEntry(Gtk.Entry, Gtk.Editable):  # type:ignore[misc]
    def do_insert_text(self, new_text, length, position):
        """Filter inserted characters to only accept numbers"""
        new_text = "".join([c for c in new_text if c.isnumeric()])
        if new_text:
            self.get_buffer().insert_text(position, new_text, length)
            return position + length
        return position


class FileChooserEntry(Gtk.Box):  # type:ignore[misc]
    """Editable entry with a file picker button"""

    max_completion_items = 15  # Maximum number of items to display in the autocompletion dropdown.

    __gsignals__ = {
        "changed": (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    def __init__(
        self,
        title=_("Select file"),
        action=Gtk.FileChooserAction.OPEN,
        text=None,
        default_path=None,
        warn_if_non_empty=False,
        warn_if_non_writable_parent=False,
        warn_if_ntfs=False,
        activates_default=False,
        shell_quoting=False,
    ):  # pylint: disable=too-many-arguments
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=0, visible=True)
        self.title = title
        self.action = action
        self.warn_if_non_empty = warn_if_non_empty
        self.warn_if_non_writable_parent = warn_if_non_writable_parent
        self.warn_if_ntfs = warn_if_ntfs
        self.shell_quoting = shell_quoting

        self.path_completion = Gtk.ListStore(str)

        self.entry = Gtk.Entry(visible=True)
        self.set_text(text)  # do before set up signal handlers
        self.original_text = self.get_text()
        self.default_path = os.path.expanduser(default_path) if default_path else self.get_path()

        self.entry.set_activates_default(activates_default)
        self.entry.set_completion(self.get_completion())
        self.entry.connect("changed", self.on_entry_changed)
        self.entry.connect("activate", self.on_activate)
        self.entry.connect("focus-out-event", self.on_focus_out)
        self.entry.connect("backspace", self.on_backspace)

        browse_button = Gtk.Button.new_from_icon_name("view-more-horizontal-symbolic", Gtk.IconSize.BUTTON)
        browse_button.show()
        if action == Gtk.FileChooserAction.SELECT_FOLDER:
            browse_button.set_tooltip_text(_("Select a folder"))
        else:
            browse_button.set_tooltip_text(_("Select a file"))
        browse_button.get_style_context().add_class("circular")
        browse_button.connect("clicked", self.on_browse_clicked)

        self.open_button = Gtk.Button.new_from_icon_name("folder-symbolic", Gtk.IconSize.BUTTON)
        self.open_button.show()
        self.open_button.set_tooltip_text(_("Open in file browser"))
        self.open_button.get_style_context().add_class("circular")
        self.open_button.connect("clicked", self.on_open_clicked)
        self.open_button.set_sensitive(bool(self.get_open_directory()))

        box = Gtk.Box(spacing=6, visible=True)
        box.pack_start(self.entry, True, True, 0)
        box.pack_end(self.open_button, False, False, 0)
        box.pack_end(browse_button, False, False, 0)
        self.pack_start(box, False, False, 0)

    def set_text(self, text):
        if self.shell_quoting and text:
            try:
                command_array = shlex.split(self.get_text())
            except ValueError:
                command_array = None  # split can fail due to imbalanced quoted

            if command_array:
                expanded = os.path.expanduser(command_array[0])
                command_array[0] = expanded
                rejoined = shlex.join(command_array)
                self.original_text = rejoined
                self.entry.set_text(rejoined)
                return

        expanded = os.path.expanduser(text) if text else ""
        self.original_text = expanded
        self.entry.set_text(expanded)

    def set_path(self, path):
        if self.shell_quoting:
            try:
                command_array = shlex.split(self.get_text())
            except ValueError:
                command_array = None  # split can fail due to imbalanced quoted

            if command_array:
                command_array[0] = os.path.expanduser(path) if path else ""
                rejoined = shlex.join(command_array)
                self.original_text = rejoined
                self.entry.set_text(rejoined)
                return

        expanded = os.path.expanduser(path) if path else ""
        self.original_text = expanded
        self.entry.set_text(expanded)

    def get_text(self):
        """Return the entry's text. If shell_quoting is one, this is actually a command
        line (with argument quoting) and not a simple path."""
        return self.entry.get_text()

    def get_path(self):
        """Returns the path in the entry; if shell_quoting is set, this extracts
        the command from the text and returns only that."""
        text = self.get_text()
        if self.shell_quoting:
            try:
                command_array = shlex.split(text)
                return command_array[0] if command_array else ""
            except ValueError:
                pass

        return text

    def get_completion(self):
        """Return an EntryCompletion widget"""
        completion = Gtk.EntryCompletion()
        completion.set_model(self.path_completion)
        completion.set_text_column(0)
        return completion

    def get_filechooser_dialog(self):
        """Return an instance of a FileChooserNative configured for this widget"""
        dialog = Gtk.FileChooserNative.new(self.title, self.get_toplevel(), self.action, _("_OK"), _("_Cancel"))
        dialog.set_create_folders(True)
        dialog.set_current_folder(self.get_default_folder())
        return dialog

    def get_default_folder(self):
        """Return the default folder for the file picker"""
        default_path = self.get_path() or self.default_path or ""
        if not default_path or not system.path_exists(default_path):
            current_entry = self.get_text()
            if system.path_exists(current_entry):
                default_path = current_entry
        if not os.path.isdir(default_path):
            default_path = os.path.dirname(default_path)
        return os.path.expanduser(default_path or "~")

    def on_browse_clicked(self, _widget):
        """Browse button click callback"""
        file_chooser_dialog = self.get_filechooser_dialog()
        response = file_chooser_dialog.run()

        if response == Gtk.ResponseType.ACCEPT:
            target_path = file_chooser_dialog.get_filename()

            if target_path and self.shell_quoting:
                try:
                    command_array = shlex.split(self.entry.get_text())
                    text = shlex.join([target_path] + command_array[1:])
                except ValueError:
                    text = shlex.join([target_path])
            else:
                text = target_path

            self.original_text = text
            self.entry.set_text(text)

        file_chooser_dialog.destroy()

    def on_open_clicked(self, _widget):
        path = self.get_open_directory()

        if path:
            open_uri(path)

    def get_open_directory(self):
        path = self.get_path()

        while path and not os.path.isdir(path):
            path = os.path.dirname(path)

        return path

    def on_entry_changed(self, widget):
        """Entry changed callback"""
        self.clear_warnings()

        # If the user isn't editing this entry, we'll apply updates
        # immediately upon any change

        if not self.entry.has_focus():
            if self.normalize_path():
                # We changed the text on commit, so we return here to avoid a double changed signal
                return

        text = self.get_text()
        path = self.get_path()
        self.original_text = text

        if self.warn_if_ntfs and LINUX_SYSTEM.get_fs_type_for_path(path) == "ntfs":
            ntfs_box = Gtk.Box(spacing=6, visible=True)
            warning_image = Gtk.Image(visible=True)
            warning_image.set_from_icon_name("dialog-warning", Gtk.IconSize.DND)
            ntfs_box.add(warning_image)
            ntfs_label = Gtk.Label(visible=True)
            ntfs_label.set_markup(
                _(
                    "<b>Warning!</b> The selected path is located on a drive formatted by Windows.\n"
                    "Games and programs installed on Windows drives <b>don't work</b>."
                )
            )
            ntfs_box.add(ntfs_label)
            self.pack_end(ntfs_box, False, False, 10)
        if self.warn_if_non_empty and os.path.exists(path) and os.listdir(path):
            non_empty_label = Gtk.Label(visible=True)
            non_empty_label.set_markup(
                _("<b>Warning!</b> The selected path contains files. Installation will not work properly.")
            )
            self.pack_end(non_empty_label, False, False, 10)
        if self.warn_if_non_writable_parent:
            parent = system.get_existing_parent(path)
            if parent is not None and not os.access(parent, os.W_OK):
                non_writable_destination_label = Gtk.Label(visible=True)
                non_writable_destination_label.set_markup(
                    _("<b>Warning</b> The destination folder is not writable by the current user.")
                )
                self.pack_end(non_writable_destination_label, False, False, 10)

        self.open_button.set_sensitive(bool(self.get_open_directory()))

        self.emit("changed")

    def on_activate(self, _widget):
        self.normalize_path()
        self.detect_changes()

    def on_focus_out(self, _widget, _event):
        self.normalize_path()
        self.detect_changes()

    def on_backspace(self, _widget):
        GLib.idle_add(self.detect_changes)

    def detect_changes(self):
        """Detects if the text has changed and updates self.original_text and fires
        the changed signal. Lame, but Gtk.Entry does not always fire its
        changed event when edited!"""
        new_text = self.get_text()
        if self.original_text != new_text:
            self.original_text = new_text
            self.emit("changed")
        return False  # used as idle function

    def normalize_path(self):
        original_path = self.get_path()
        path = original_path.strip("\r\n")

        if path.startswith("file:///"):
            path = urllib.parse.unquote(path[len("file://") :])

        path = os.path.expanduser(path)

        self.update_completion(path)

        if path != original_path:
            self.set_path(path)
            return True

        return False

    def update_completion(self, current_path):
        """Update the auto-completion widget with the current path"""
        self.path_completion.clear()

        if not os.path.exists(current_path):
            current_path, filefilter = os.path.split(current_path)
        else:
            filefilter = None

        if os.path.isdir(current_path):
            index = 0
            for filename in sorted(os.listdir(current_path)):
                if filename.startswith("."):
                    continue
                if filefilter is not None and not filename.startswith(filefilter):
                    continue
                self.path_completion.append([os.path.join(current_path, filename)])
                index += 1
                if index > self.max_completion_items:
                    break

    def clear_warnings(self):
        """Delete all the warning labels from the container"""
        for index, child in enumerate(self.get_children()):
            if index > 0:
                child.destroy()


class Label(Gtk.Label):
    """Standardised label for config vboxes."""

    def __init__(self, message=None, width_request=230, max_width_chars=22, visible=True):
        """Custom init of label."""
        super().__init__(label=message, visible=visible)
        self.set_line_wrap(True)
        self.set_max_width_chars(max_width_chars)
        self.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
        self.set_size_request(width_request, -1)
        self.set_alignment(0, 0.5)
        self.set_justify(Gtk.Justification.LEFT)


class VBox(Gtk.Box):
    def __init__(self, **kwargs):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, margin_top=18, **kwargs)


class EditableGrid(Gtk.Box):
    __gsignals__ = {"changed": (GObject.SIGNAL_RUN_FIRST, None, ())}

    def __init__(self, data, columns):
        self.columns = columns
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=6)

        self.liststore = Gtk.ListStore(str, str)
        for item in data:
            self.liststore.append([str(value) for value in item])

        self.treeview = Gtk.TreeView.new_with_model(self.liststore)
        self.treeview.set_grid_lines(Gtk.TreeViewGridLines.BOTH)
        for i, column_title in enumerate(self.columns):
            renderer = Gtk.CellRendererText()
            renderer.set_property("editable", True)
            renderer.connect("edited", self.on_text_edited, i)

            column = Gtk.TreeViewColumn(column_title, renderer, text=i)
            column.set_resizable(True)
            column.set_min_width(100)
            column.set_sort_column_id(0)
            self.treeview.append_column(column)

        self.buttons = []
        self.add_button = Gtk.Button(_("Add"))
        self.buttons.append(self.add_button)
        self.add_button.connect("clicked", self.on_add)

        self.delete_button = Gtk.Button(_("Delete"))
        self.buttons.append(self.delete_button)
        self.delete_button.connect("clicked", self.on_delete)

        self.scrollable_treelist = Gtk.ScrolledWindow()
        self.scrollable_treelist.set_size_request(-1, 209)
        self.scrollable_treelist.add(self.treeview)
        self.scrollable_treelist.set_shadow_type(Gtk.ShadowType.IN)

        self.pack_start(self.scrollable_treelist, True, True, 0)
        button_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        for button in reversed(self.buttons):
            button.set_size_request(80, -1)
            button_box.pack_end(button, False, False, 0)
        self.pack_end(button_box, False, False, 0)
        self.show_all()

    def on_add(self, widget):  # pylint: disable=unused-argument
        self.liststore.append(["", ""])
        row_position = len(self.liststore) - 1
        self.treeview.set_cursor(row_position, None, False)
        self.treeview.scroll_to_cell(row_position, None, False, 0.0, 0.0)
        self.emit("changed")

    def on_delete(self, widget):  # pylint: disable=unused-argument
        selection = self.treeview.get_selection()
        _, iteration = selection.get_selected()
        if iteration:
            self.liststore.remove(iteration)
            self.emit("changed")

    def on_text_edited(self, widget, path, text, field):  # pylint: disable=unused-argument
        self.liststore[path][field] = text.strip()  # pylint: disable=unsubscriptable-object
        self.emit("changed")

    def get_data(self):  # pylint: disable=arguments-differ
        model_data = []
        for row in self.liststore:  # pylint: disable=not-an-iterable
            model_data.append(row)
        return model_data



from gi.repository import GObject, Gtk

from lutris.gui.installer.script_box import InstallerScriptBox


class InstallerPicker(Gtk.ListBox):
    """List box to pick between several installers"""

    __gsignals__ = {"installer-selected": (GObject.SIGNAL_RUN_FIRST, None, (str,))}

    def __init__(self, scripts):
        super().__init__()
        revealed = True
        for script in scripts:
            self.add(InstallerScriptBox(script, parent=self, revealed=revealed))
            revealed = False  # Only reveal the first installer.
        self.connect("row-selected", self.on_activate)
        self.show_all()

    @staticmethod
    def on_activate(widget, row):
        """Handler for hiding and showing the revealers in children"""
        for script_box_row in widget:
            script_box = script_box_row.get_children()[0]
            script_box.reveal(False)
        installer_row = row.get_children()[0]
        installer_row.reveal()



from collections import OrderedDict
from copy import deepcopy
from gettext import gettext as _

from gi.repository import Gio, GLib, Gtk

from lutris.config import write_game_config
from lutris.database.games import add_game
from lutris.game import GAME_INSTALLED, GAME_UPDATED, Game
from lutris.gui.dialogs import ModelessDialog
from lutris.scanners.default_installers import DEFAULT_INSTALLERS
from lutris.scanners.tosec import clean_rom_name, guess_platform, search_tosec_by_md5
from lutris.services.lutris import download_lutris_media
from lutris.util.jobs import AsyncCall
from lutris.util.log import logger
from lutris.util.path_cache import get_path_cache
from lutris.util.strings import gtk_safe, slugify
from lutris.util.system import get_md5_hash, get_md5_in_zip


class ImportGameDialog(ModelessDialog):
    def __init__(self, files, parent=None) -> None:
        super().__init__(_("Import a game"), parent=parent, border_width=10)
        self.files = files
        self.progress_labels = {}
        self.checksum_labels = {}
        self.description_labels = {}
        self.category_labels = {}
        self.error_labels = {}
        self.launch_buttons = {}
        self.platform = None
        self.search_call = None
        self.set_size_request(500, 560)

        self.accelerators = Gtk.AccelGroup()
        self.add_accel_group(self.accelerators)

        scrolledwindow = Gtk.ScrolledWindow(child=self.get_file_labels_listbox(files))
        scrolledwindow.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        frame = Gtk.Frame(shadow_type=Gtk.ShadowType.ETCHED_IN, child=scrolledwindow)
        self.get_content_area().pack_start(frame, True, True, 6)

        self.close_button = self.add_button(Gtk.STOCK_STOP, Gtk.ResponseType.CANCEL)
        key, mod = Gtk.accelerator_parse("Escape")
        self.close_button.add_accelerator("clicked", self.accelerators, key, mod, Gtk.AccelFlags.VISIBLE)

        self.show_all()
        self.search_call = AsyncCall(self.search_checksums, self.search_result_finished)

    def on_response(self, dialog, response: Gtk.ResponseType) -> None:
        if response in (Gtk.ResponseType.CLOSE, Gtk.ResponseType.CANCEL, Gtk.ResponseType.DELETE_EVENT):
            if self.search_call:
                self.search_call.stop_request.set()
                return  # don't actually close the dialog

        super().on_response(dialog, response)

    def get_file_labels_listbox(self, files):
        listbox = Gtk.ListBox(vexpand=True)
        listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        for file_path in files:
            row = Gtk.ListBoxRow()
            hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
            hbox.set_margin_left(12)
            hbox.set_margin_right(12)

            vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

            description_label = Gtk.Label(halign=Gtk.Align.START)
            vbox.pack_start(description_label, True, True, 5)
            self.description_labels[file_path] = description_label

            file_path_label = Gtk.Label(file_path, halign=Gtk.Align.START, xalign=0)
            file_path_label.set_line_wrap(True)
            vbox.pack_start(file_path_label, True, True, 5)

            progress_label = Gtk.Label(halign=Gtk.Align.START)
            vbox.pack_start(progress_label, True, True, 5)
            self.progress_labels[file_path] = progress_label

            checksum_label = Gtk.Label(no_show_all=True, halign=Gtk.Align.START)
            vbox.pack_start(checksum_label, True, True, 5)
            self.checksum_labels[file_path] = checksum_label

            category_label = Gtk.Label(no_show_all=True, halign=Gtk.Align.START)
            vbox.pack_start(category_label, True, True, 5)
            self.category_labels[file_path] = category_label

            error_label = Gtk.Label(no_show_all=True, halign=Gtk.Align.START, xalign=0)
            error_label.set_line_wrap(True)
            vbox.pack_start(error_label, True, True, 5)
            self.error_labels[file_path] = error_label

            hbox.pack_start(vbox, True, True, 0)

            launch_button = Gtk.Button(_("Launch"), valign=Gtk.Align.CENTER, sensitive=False)
            hbox.pack_end(launch_button, False, False, 0)
            self.launch_buttons[file_path] = launch_button

            row.add(hbox)
            listbox.add(row)
        return listbox

    @property
    def search_stopping(self):
        return self.search_call and self.search_call.stop_request.is_set()

    def search_checksums(self):
        game_path_cache = get_path_cache()

        def show_progress(filepath, message):
            # It's not safe to directly update labels from a worker thread, so
            # this will do it on the GUI main thread instead.
            GLib.idle_add(lambda: self.progress_labels[filepath].set_markup("<i>%s</i>" % gtk_safe(message)))

        def get_existing_game(filepath):
            for game_id, game_path in game_path_cache.items():
                if game_path == filepath:
                    return Game(game_id)

            return None

        def search_single(filepath):
            existing_game = get_existing_game(filepath)
            if existing_game:
                # Found a game to launch instead of installing, but we can't safely
                # do this on this thread, so we return the game and handle it later.
                return [{"name": existing_game.name, "game": existing_game, "roms": []}]
            show_progress(filepath, _("Calculating checksum..."))
            if filepath.lower().endswith(".zip"):
                md5 = get_md5_in_zip(filepath)
            else:
                md5 = get_md5_hash(filepath)

            if self.search_stopping:
                return None

            show_progress(filename, _("Looking up checksum on Lutris.net..."))
            result = search_tosec_by_md5(md5)
            if not result:
                raise RuntimeError(_("This ROM could not be identified."))
            return result

        results = OrderedDict()  # must preserve order, on any Python version
        for filename in self.files:
            if self.search_stopping:
                break

            try:
                show_progress(filename, _("Looking for installed game..."))
                result = search_single(filename)
            except Exception as error:
                result = [{"error": error, "roms": []}]
            finally:
                show_progress(filename, "")

            if result:
                results[filename] = result

        return results

    def search_result_finished(self, results, error):
        self.search_call = None
        self.close_button.set_label(Gtk.STOCK_CLOSE)

        if error:
            logger.error(error)
            return

        for filename, result in results.items():
            for rom_set in result:
                if self.import_rom(rom_set, filename):
                    break

    def import_rom(self, rom_set, filename):
        """Tries to install a specific ROM, or reports failure. Returns True if
        successful, False if not."""
        try:
            self.progress_labels[filename].hide()

            if "error" in rom_set:
                raise rom_set["error"]

            if "game" in rom_set:
                game = rom_set["game"]
                self.display_existing_game_info(filename, game)
                self.enable_game_launch(filename, game)
                return True

            for rom in rom_set["roms"]:
                self.display_new_game_info(filename, rom_set, rom["md5"])
                game_id = self.add_game(rom_set, filename)
                game = Game(game_id)
                GAME_INSTALLED.fire(game)
                GAME_UPDATED.fire(game)
                self.enable_game_launch(filename, game)
                return True
        except Exception as ex:
            logger.exception(_("Failed to import a ROM: %s"), ex)
            error_label = self.error_labels[filename]
            error_label.set_markup('<span style="italic" foreground="red">%s</span>' % gtk_safe(str(ex)))
            error_label.show()

        return False

    def enable_game_launch(self, filename, game):
        launch_button = self.launch_buttons[filename]
        launch_button.set_sensitive(True)
        launch_button.connect("clicked", self.on_launch_clicked, game)

    def on_launch_clicked(self, _button, game):
        # We can't use this window as the delegate because we
        # are destroying it.
        application = Gio.Application.get_default()
        game.launch(application.launch_ui_delegate)
        self.destroy()

    def display_existing_game_info(self, filename, game):
        label = self.checksum_labels[filename]
        label.set_markup("<i>%s</i>" % _("Game already installed in Lutris"))
        label.show()
        label = self.description_labels[filename]
        label.set_markup("<b>%s</b>" % game.name)
        category = game.platform
        label = self.category_labels[filename]
        label.set_text(category)
        label.show()

    def display_new_game_info(self, filename, rom_set, checksum):
        label = self.checksum_labels[filename]
        label.set_text(checksum)
        label.show()
        label = self.description_labels[filename]
        label.set_markup("<b>%s</b>" % rom_set["name"])
        category = rom_set["category"]["name"]
        label = self.category_labels[filename]
        label.set_text(category)
        label.show()
        self.platform = guess_platform(rom_set)

        if not self.platform:
            raise RuntimeError(_("The platform '%s' is unknown to Lutris.") % category)

    def add_game(self, rom_set, filepath):
        name = clean_rom_name(rom_set["name"])
        logger.info("Installing %s", name)

        try:
            installer = deepcopy(DEFAULT_INSTALLERS[self.platform])
        except KeyError as error:
            raise RuntimeError(
                _("Lutris does not have a default installer for the '%s' platform.") % self.platform
            ) from error

        for key, value in installer["game"].items():
            if value == "rom":
                installer["game"][key] = filepath
        slug = slugify(name)
        configpath = write_game_config(slug, installer)
        game_id = add_game(
            name=name,
            runner=installer["runner"],
            slug=slug,
            directory="",
            installed=1,
            installer_slug="%s-%s" % (slug, installer["runner"]),
            configpath=configpath,
        )
        download_lutris_media(slug)
        return game_id



import time
from http.cookiejar import Cookie, MozillaCookieJar


class WebkitCookieJar(MozillaCookieJar):
    """Subclass of MozillaCookieJar for compatibility with cookies
    coming from Webkit2.
    This disables the magic_re header which is not present and adds
    compatibility with HttpOnly cookies (See http://bugs.python.org/issue2190)
    """

    def _really_load(self, f, filename, ignore_discard, ignore_expires):  # pylint: disable=too-many-locals
        now = time.time()
        try:
            while 1:
                line = f.readline()
                if line == "":
                    break

                # last field may be absent, so keep any trailing tab
                if line.endswith("\n"):
                    line = line[:-1]

                sline = line.strip()
                # support HttpOnly cookies (as stored by curl or old Firefox).
                if sline.startswith("#HttpOnly_"):
                    line = sline[10:]
                elif sline.startswith("#") or sline == "":
                    continue

                domain, domain_specified, path, secure, expires, name, value, *_extra = line.split("\t")
                secure = secure == "TRUE"
                domain_specified = domain_specified == "TRUE"
                if name == "":
                    # cookies.txt regards 'Set-Cookie: foo' as a cookie
                    # with no name, whereas http.cookiejar regards it as a
                    # cookie with no value.
                    name = value
                    value = None

                initial_dot = domain.startswith(".")
                assert domain_specified == initial_dot

                discard = False
                if expires == "":
                    expires = None
                    discard = True

                # assume path_specified is false
                c = Cookie(
                    0,
                    name,
                    value,
                    None,
                    False,
                    domain,
                    domain_specified,
                    initial_dot,
                    path,
                    False,
                    secure,
                    expires,
                    discard,
                    None,
                    None,
                    {},
                )
                if not ignore_discard and c.discard:
                    continue
                if not ignore_expires and c.is_expired(now):
                    continue
                self.set_cookie(c)

        except OSError:
            raise
        except Exception as err:
            raise OSError("invalid Netscape format cookies file %r: %r" % (filename, line)) from err



"""Steam Family service"""

import json
import os
from gettext import gettext as _

import requests

from lutris import settings
from lutris.services.base import SERVICE_LOGIN, AuthTokenExpiredError, OnlineService
from lutris.services.steam import SteamGame, SteamService
from lutris.util.log import logger
from lutris.util.steam.config import get_active_steamid64, get_steam_library


class SteamFamilyGame(SteamGame):
    service = "steamfamily"
    installer_slug = "steam"
    runner = "steam"

    @classmethod
    def new_from_steamfamily_game(cls, game):
        """Return a Steam Family game instance from an AppManifest"""
        game = cls.new_from_steam_game(game)
        game.service = cls.service
        return game


class SteamFamilyService(SteamService, OnlineService):
    """Service class for Steam Family sharing"""

    id = "steamfamily"
    name = _("Steam Family")
    description = _("Use for displaying every game in the Steam family")
    login_window_width = 500
    login_window_height = 850
    online = True
    requires_login_page = True
    game_class = SteamFamilyGame
    include_own_games = settings.STEAM_FAMILY_INCLUDE_OWN
    cookies_path = os.path.join(settings.CACHE_DIR, ".steam.auth")
    token_path = os.path.join(settings.CACHE_DIR, ".steam.token")
    cache_path = os.path.join(settings.CACHE_DIR, "steam-library.json")
    login_url = "https://store.steampowered.com/login/?redir=/about"
    redirect_uris = ["https://store.steampowered.com/about"]
    access_token_url = "https://store.steampowered.com/pointssummary/ajaxgetasyncconfig"
    library_url = "https://api.steampowered.com/IFamilyGroupsService/GetSharedLibraryApps/v1/"
    family_url = "https://api.steampowered.com/IFamilyGroupsService/GetFamilyGroupForUser/v1/"

    user_agent = (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
        "AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/84.0.4147.38 Safari/537.36"
    )

    def __init__(self):
        super().__init__()
        self.session = requests.session()
        self.session.headers["User-Agent"] = self.user_agent
        self.access_token = self.load_access_token()

    def is_connected(self):
        return self.is_authenticated() and bool(self.load_access_token())

    def fetch_access_token(self):
        """Fetch the access token from the store, save to disk and set"""
        token_data = self.get_access_token()
        if not token_data:
            raise RuntimeError("Failed to get access token")
        with open(self.token_path, "w", encoding="utf-8") as token_file:
            token_file.write(json.dumps(token_data, indent=2))
        self.access_token = self.load_access_token()

    def load_access_token(self):
        """Load the access token from disk"""
        if not os.path.exists(self.token_path):
            return ""
        with open(self.token_path, encoding="utf-8") as token_file:
            token_data = json.load(token_file)
            return token_data.get("data").get("webapi_token", "")

    def get_access_token(self):
        """Request an access token from steam and return dump"""
        logger.debug("Requesting access token")
        response = self.session.get(
            self.access_token_url,
            cookies=self.load_cookies(),
        )
        response.raise_for_status()
        token_data = response.json()
        return token_data

    def login_callback(self, content):
        """Once the user logs in in a browser window, they're redirected to a
        an arbitrary page (about), then we redirect to a page containing the
        store access token which we can use to fetch family games"""
        logger.debug("Login to Steam store successful")
        self.fetch_access_token()
        SERVICE_LOGIN.fire(self)

    def get_family_groupid(self):
        """Get the user's family group id"""
        response = self.session.get(
            self.family_url, params={"access_token": self.load_access_token(), "steamid": get_active_steamid64()}
        )
        response.raise_for_status()
        resData = response.json()
        records = resData["response"]
        if not records["is_not_member_of_any_group"]:
            return records["family_groupid"]
        logger.error("User is not a member of any family group")
        return None

    def get_library(self):
        steamid = get_active_steamid64()
        if not steamid:
            logger.error("Unable to find SteamID from Steam config")
            return []
        response = self.session.get(
            self.library_url,
            params={
                "access_token": self.load_access_token(),
                "family_groupid": self.get_family_groupid(),
                "steamid": get_active_steamid64(),
            },
        )
        response.raise_for_status()
        resData = response.json()
        records = resData["response"]["apps"]
        if self.include_own_games:
            own_games = get_steam_library(steamid)
            ids = {game["appid"] for game in records}
            records.extend([game for game in own_games if game["appid"] not in ids])
        return records

    def load(self):
        """Load the list of games"""
        try:
            library = self.get_library()
        except Exception as ex:  # pylint=disable:broad-except
            logger.warning("Access Token expired, will attempt to get a new one")
            try:
                self.fetch_access_token()
                library = self.get_library()
            except:
                logger.warning("Failed to get a new access token")
                raise AuthTokenExpiredError("Access Token expired") from ex
        for steam_game in library:
            if steam_game["appid"] in self.excluded_appids:
                continue
            game = self.game_class.new_from_steamfamily_game(steam_game)
            game.save()
        self.match_games()
        return library



from gettext import gettext as _
from typing import Dict, Iterable, List, Tuple, Union

from gi.repository import Gdk, Gtk

from lutris.gui.config.base_config_box import BaseConfigBox
from lutris.gui.widgets.log_text_view import LogTextView
from lutris.util import linux, system
from lutris.util.linux import gather_system_info_dict
from lutris.util.log import get_log_contents
from lutris.util.strings import gtk_safe
from lutris.util.wine.wine import is_esync_limit_set, is_fsync_supported, is_installed_systemwide


class SystemBox(BaseConfigBox):
    features_definitions = [
        {
            "name": _("Vulkan support"),
            "callable": linux.LINUX_SYSTEM.is_vulkan_supported,
        },
        {
            "name": _("Esync support"),
            "callable": is_esync_limit_set,
        },
        {
            "name": _("Fsync support"),
            "callable": is_fsync_supported,
        },
        {
            "name": _("Wine installed"),
            "callable": is_installed_systemwide,
        },
        {"name": _("Gamescope"), "callable": system.can_find_executable, "args": ("gamescope",)},
        {"name": _("Mangohud"), "callable": system.can_find_executable, "args": ("mangohud",)},
        {"name": _("Gamemode"), "callable": linux.LINUX_SYSTEM.gamemode_available},
        {"name": _("Steam"), "callable": linux.LINUX_SYSTEM.has_steam},
        {"name": _("In Flatpak"), "callable": linux.LINUX_SYSTEM.is_flatpak},
    ]

    def __init__(self):
        super().__init__()
        self.pack_start(self.get_section_label(_("System information")), False, False, 0)

        self.scrolled_window = Gtk.ScrolledWindow(visible=True)
        self.scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sysinfo_frame = Gtk.Frame(visible=True)
        sysinfo_frame.get_style_context().add_class("info-frame")
        sysinfo_frame.add(self.scrolled_window)
        self.pack_start(sysinfo_frame, True, True, 0)

        button_copy = Gtk.Button(_("Copy system info to Clipboard"), halign=Gtk.Align.START, visible=True)
        button_copy.connect("clicked", self.on_copy_clicked)

        self.pack_start(button_copy, False, False, 0)

        self.pack_start(self.get_section_label(_("Lutris logs")), False, False, 0)

        self.log_scrolled_window = Gtk.ScrolledWindow(visible=True)
        self.log_scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        log_frame = Gtk.Frame(visible=True)
        log_frame.get_style_context().add_class("info-frame")
        log_frame.add(self.log_scrolled_window)
        self.pack_start(log_frame, True, True, 0)

        button_log_copy = Gtk.Button(_("Copy logs to Clipboard"), halign=Gtk.Align.START, visible=True)
        button_log_copy.connect("clicked", self.on_copy_log_clicked)

        self.pack_start(button_log_copy, False, False, 0)

    def populate(self):
        items = self.get_items()
        self.scrolled_window.add(self.get_grid(items))
        self.log_scrolled_window.add(self.get_log_view())

    def get_log_view(self):
        log_buffer = Gtk.TextBuffer()
        log_buffer.set_text(get_log_contents())
        return LogTextView(log_buffer)

    def get_items(self) -> list:
        """Assembles a list of items to display; most items are name-value tuples
        giving various bits of information, section headers appear also, as plain strings."""
        features = self.get_features()
        items: List[Union[str, Tuple[str, str]]] = [(f["name"], f["available_text"]) for f in features]

        system_info_readable = gather_system_info_dict()
        for section, dictionary in system_info_readable.items():
            items.append(section)
            items.extend(dictionary.items())

        return items

    @staticmethod
    def get_grid(items: Iterable) -> Gtk.Grid:
        """Constructs a Gtk.Grid containing labels for each item given; each item
        may be a name-value tuple, producing two labels, or just a string, giving one
        that covers two columns; this later is used for section headers."""
        grid = Gtk.Grid(visible=True, row_spacing=6, margin=16)
        row = 0
        for item in items:
            if isinstance(item, str):
                header_label = Gtk.Label(visible=True, xalign=0, yalign=0, margin_top=16)
                header_label.set_markup("<b>[%s]</b>" % gtk_safe(item))
                if row == 0:
                    grid.set_margin_top(0)
                grid.attach(header_label, 0, row, 2, 1)
            else:
                name, text = item
                name_label = Gtk.Label(name + ":", visible=True, xalign=0, yalign=0, margin_right=30)
                grid.attach(name_label, 0, row, 1, 1)

                markup_label = Gtk.Label(visible=True, xalign=0, selectable=True)
                markup_label.set_markup("<b>%s</b>" % gtk_safe(text))
                grid.attach(markup_label, 1, row, 1, 1)
            row += 1
        return grid

    @staticmethod
    def get_text(items: Iterable) -> str:
        """Constructs text for the clipboard, given the same items as get_grid() takess"""
        lines = []
        for item in items:
            if isinstance(item, str):
                lines.append(f"[{item}]")
            else:
                name, text = item
                lines.append(f"{name}: {text}")
        return "\n".join(lines)

    def get_features(self) -> List[Dict[str, str]]:
        """Provides a list of features that may be present in your system; each
        is given as a dict, which hase 'name' and 'available_text' keys."""
        yes = _("YES")
        no = _("NO")

        def eval_feature(feature):
            result = feature.copy()
            func = feature["callable"]
            args = feature.get("args", ())
            result["availability"] = bool(func(*args))
            result["available_text"] = yes if result["availability"] else no
            return result

        return [eval_feature(f) for f in self.features_definitions]

    def on_copy_clicked(self, _widget) -> None:
        items = self.get_items()
        text = self.get_text(items)

        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(text.strip(), -1)

    def on_copy_log_clicked(self, _widget) -> None:
        text = get_log_contents()
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(text.strip(), -1)



"""Misc common functions"""

import functools
from typing import Any, Callable, Mapping

AnyCallable = Callable[..., Any]


def selective_merge(base_obj: Any, delta_obj: Mapping[Any, Any]) -> Mapping[Any, Any]:
    """used by write_json"""
    if not isinstance(base_obj, dict):
        return delta_obj
    common_keys = set(base_obj).intersection(delta_obj)
    new_keys = set(delta_obj).difference(common_keys)
    for k in common_keys:
        base_obj[k] = selective_merge(base_obj[k], delta_obj[k])
    for k in new_keys:
        base_obj[k] = delta_obj[k]
    return base_obj


class cache_single:
    """A simple replacement for lru_cache, with no LRU behavior. This caches
    a single result from a function that has no arguments at all. Exceptions
    are not cached; there's a 'clear_cache()' function on the wrapper like with
    lru_cache to explicitly clear the cache."""

    def __init__(self, function: AnyCallable):
        functools.update_wrapper(self, function)
        self.function = function
        self.is_cached = False
        self.cached_item = None

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if args or kwargs:
            return self.function(*args, **kwargs)

        if not self.is_cached:
            self.cached_item = self.function()
            self.is_cached = True

        return self.cached_item

    def cache_clear(self) -> None:
        self.is_cached = False
        self.cached_item = None



# Standard Library
import os
import re
import shutil
from collections import Counter
from configparser import RawConfigParser
from gettext import gettext as _

# Lutris Modules
from lutris import settings
from lutris.runners.runner import Runner
from lutris.util import joypad, system


class reicast(Runner):
    human_name = _("Reicast")
    description = _("Sega Dreamcast emulator")
    platforms = [_("Sega Dreamcast")]
    runner_executable = "reicast/reicast.elf"
    entry_point_option = "iso"
    flatpak_id = "org.flycast.Flycast"
    joypads = None

    game_options = [
        {
            "option": "iso",
            "type": "file",
            "label": _("Disc image file"),
            "help": _("The game data.\nSupported formats: ISO, CDI"),
        }
    ]

    def __init__(self, config=None):
        super().__init__(config)

        self.runner_options = [
            {
                "option": "fullscreen",
                "type": "bool",
                "label": _("Fullscreen"),
                "default": False,
            },
            {
                "option": "device_id_1",
                "type": "choice",
                "section": _("Gamepads"),
                "label": _("Gamepad 1"),
                "choices": self.get_joypads,
                "default": "-1",
            },
            {
                "option": "device_id_2",
                "type": "choice",
                "section": _("Gamepads"),
                "label": _("Gamepad 2"),
                "choices": self.get_joypads,
                "default": "-1",
            },
            {
                "option": "device_id_3",
                "type": "choice",
                "section": _("Gamepads"),
                "label": _("Gamepad 3"),
                "choices": self.get_joypads,
                "default": "-1",
            },
            {
                "option": "device_id_4",
                "type": "choice",
                "section": _("Gamepads"),
                "label": _("Gamepad 4"),
                "choices": self.get_joypads,
                "default": "-1",
            },
        ]

    def install(self, install_ui_delegate, version=None, callback=None):
        def on_runner_installed(*args):
            mapping_path = system.create_folder("~/.reicast/mappings")
            mapping_source = os.path.join(settings.RUNNER_DIR, "reicast/mappings")
            for mapping_file in os.listdir(mapping_source):
                shutil.copy(os.path.join(mapping_source, mapping_file), mapping_path)
            system.create_folder("~/.reicast/data")

        super().install(install_ui_delegate, version, on_runner_installed)

    def get_joypads(self):
        """Return list of joypad in a format usable in the options"""
        if self.joypads:
            return self.joypads
        joypad_list = [("No joystick", "-1")]
        joypad_devices = joypad.get_joypads()
        name_counter = Counter([j[1] for j in joypad_devices])
        name_indexes = {}
        for dev, joy_name in joypad_devices:
            dev_id = re.findall(r"(\d+)", dev)[0]
            if name_counter[joy_name] > 1:
                if joy_name not in name_indexes:
                    index = 1
                else:
                    index = name_indexes[joy_name] + 1
                name_indexes[joy_name] = index
            else:
                index = 0
            if index:
                joy_name += " (%d)" % index
            joypad_list.append((joy_name, dev_id))
        self.joypads = joypad_list
        return joypad_list

    @staticmethod
    def write_config(config):
        # use RawConfigParser to preserve case-sensitive configs written by Reicast
        # otherwise, Reicast will write with mixed-case and Lutris will overwrite with all lowercase
        #   which will confuse Reicast
        parser = RawConfigParser()
        parser.optionxform = lambda option: option

        config_path = os.path.expanduser("~/.reicast/emu.cfg")
        if system.path_exists(config_path):
            with open(config_path, "r", encoding="utf-8") as config_file:
                parser.read_file(config_file)

        for section in config:
            if not parser.has_section(section):
                parser.add_section(section)
            for key, value in config[section].items():
                parser.set(section, key, str(value))

        with open(config_path, "w", encoding="utf-8") as config_file:
            parser.write(config_file)

    def play(self):
        fullscreen = "1" if self.runner_config.get("fullscreen") else "0"
        reicast_config = {
            "x11": {"fullscreen": fullscreen},
            "input": {},
            "players": {"nb": "1"},
        }
        players = 1
        reicast_config["input"] = {}
        for index in range(1, 5):
            config_string = "device_id_%d" % index
            joy_id = self.runner_config.get(config_string) or "-1"
            reicast_config["input"]["evdev_{}".format(config_string)] = joy_id
            if index > 1 and joy_id != "-1":
                players += 1
        reicast_config["players"]["nb"] = players

        self.write_config(reicast_config)

        iso = self.game_config.get("iso")
        return {"command": self.get_command() + ["-config", f"config:image={iso}"]}



"""DOSBox installer commands"""

# Standard Library
import os

# Lutris Modules
from lutris import runtime
from lutris.runners import import_runner
from lutris.util import system
from lutris.util.log import logger


def dosexec(config_file=None, executable=None, args=None, close_on_exit=True, working_dir=None):
    """Execute Dosbox with given config_file."""
    if config_file:
        run_with = "config {}".format(config_file)
        if not working_dir:
            working_dir = os.path.dirname(config_file)
    elif executable:
        run_with = "executable {}".format(executable)
        if not working_dir:
            working_dir = os.path.dirname(executable)
    else:
        raise ValueError("Neither a config file or an executable were provided")
    logger.debug("Running dosbox with %s", run_with)
    working_dir = system.create_folder(working_dir)
    dosbox = import_runner("dosbox")
    dosbox_runner = dosbox()
    command = [dosbox_runner.get_executable()]
    if config_file:
        command += ["-conf", config_file]
    if executable:
        if not system.path_exists(executable):
            raise OSError("Can't find file {}".format(executable))
        command += [executable]
    if args:
        command += args.split()
    if close_on_exit:
        command.append("-exit")
    system.execute(command, cwd=working_dir, env=runtime.get_env())


def makeconfig(path, drives, commands):
    system.create_folder(os.path.dirname(path))
    with open(path, "w", encoding="utf-8") as config_file:
        config_file.write("[autoexec]\n")
        for drive in drives:
            config_file.write('mount {} "{}"\n'.format(drive, drives[drive]))
        for command in commands:
            config_file.write("{}\n".format(command))



"""Dolphin runner"""

from gettext import gettext as _

from lutris.exceptions import MissingGameExecutableError

# Lutris Modules
from lutris.runners.runner import Runner
from lutris.util import system

PLATFORMS = [_("Nintendo GameCube"), _("Nintendo Wii")]


class dolphin(Runner):
    description = _("GameCube and Wii emulator")
    human_name = _("Dolphin")
    platforms = PLATFORMS
    require_libs = [
        "libOpenGL.so.0",
    ]
    runnable_alone = True
    runner_executable = "dolphin/Dolphin_Emulator-2512-anylinux-x86_64.AppImage"
    flatpak_id = "org.DolphinEmu.dolphin-emu"
    download_url = "https://github.com/pkgforge-dev/Dolphin-emu-AppImage/releases/download/2512%402026-01-26_1769467304/Dolphin_Emulator-2512-anylinux-x86_64.AppImage"
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "default_path": "game_path",
            "label": _("ISO file"),
        },
        {
            "option": "platform",
            "type": "choice",
            "label": _("Platform"),
            "choices": ((_("Nintendo GameCube"), "0"), (_("Nintendo Wii"), "1")),
        },
    ]
    runner_options = [
        {
            "option": "batch",
            "type": "bool",
            "label": _("Batch"),
            "default": True,
            "advanced": True,
            "help": _("Exit Dolphin with emulator."),
        },
        {
            "option": "user_directory",
            "type": "directory",
            "warn_if_non_writable_parent": True,
            "advanced": True,
            "label": _("Custom Global User Directory"),
        },
    ]

    def get_platform(self):
        selected_platform = self.game_config.get("platform")
        if selected_platform:
            return self.platforms[int(selected_platform)]
        return ""

    def play(self):
        command = self.get_command()

        # Batch isn't available in nogui
        if self.runner_config.get("batch"):
            command.append("--batch")

        # Custom Global User Directory
        if self.runner_config.get("user_directory"):
            command.append("-u")
            command.append(self.runner_config["user_directory"])

        # Retrieve the path to the file
        iso = self.game_config.get("main_file") or ""
        if not system.path_exists(iso):
            raise MissingGameExecutableError(filename=iso)
        command.extend(["-e", iso])

        return {"command": command}



"""Migrate Proton versions from the old runners/proton directory to runners/wine"""

import os

from lutris import settings
from lutris.util.log import logger


def migrate():
    src_dir = os.path.join(settings.RUNNER_DIR, "proton")
    dest_dir = settings.WINE_DIR

    if not os.path.isdir(src_dir):
        return

    try:
        os.makedirs(dest_dir, exist_ok=True)

        for entry in os.listdir(src_dir):
            src_path = os.path.join(src_dir, entry)
            dest_path = os.path.join(dest_dir, entry)

            if not os.path.isdir(src_path):
                continue

            if os.path.exists(dest_path):
                logger.info("Skipping '%s', already exists in %s", entry, dest_dir)
            else:
                logger.info("Moving Proton version '%s' to %s", entry, dest_dir)
                os.rename(src_path, dest_path)

        if not os.listdir(src_dir):
            os.rmdir(src_dir)
    except OSError as ex:
        logger.exception("Failed to migrate Proton versions: %s", ex)



"""Parser for the glxinfo utility"""

from lutris.util.log import logger
from lutris.util.system import read_process_output


class Container:  # pylint: disable=too-few-public-methods
    """A dummy container for data"""


class GlxInfo:
    """Give access to the glxinfo information"""

    def __init__(self, output=None):
        """Creates a new GlxInfo object

        Params:
            output (str): If provided, use this as the glxinfo output instead
                          of running the program, useful for testing.
        """
        self._output = output or self.get_glxinfo_output()
        self._section = None
        self._attrs = set()  # Keep a reference of the created attributes
        self.parse()

    @staticmethod
    def get_glxinfo_output():
        """Return the glxinfo -B output"""
        return read_process_output(["glxinfo", "-B"])

    def as_dict(self):
        """Return the attributes as a dict"""
        return {attr: getattr(self, attr) for attr in self._attrs}

    def parse(self):
        """Converts the glxinfo output to class attributes"""
        if not self._output:
            logger.error("No available glxinfo output")
            return
        # Fix glxinfo output (Great, you saved one line by
        # combining display and screen)
        output = self._output.replace("  screen", "\nscreen")
        for line in output.split("\n"):
            # Skip blank lines, and error lines that may contain no ':'
            if ":" not in line:
                continue

            key, value = line.split(":", 1)
            key = key.replace(" string", "").replace(" ", "_")
            value = value.strip()

            if not value and key.startswith(("Extended_renderer_info", "Memory_info")) and "(" in key:
                self._section = key[key.index("(") + 1 : -1]
                setattr(self, self._section, Container())
                continue
            if self._section:
                if not key.startswith("____"):
                    self._section = None
                else:
                    setattr(getattr(self, self._section), key.strip("_").lower(), value)
                    continue
            self._attrs.add(key.lower())
            setattr(self, key.lower(), value)



"""itch.io service"""

import datetime
import json
import os
from gettext import gettext as _
from typing import Any, Dict, List, Optional
from urllib.parse import quote_plus, urlencode

from gi.repository import Gtk

from lutris import settings
from lutris.database import games as games_db
from lutris.exceptions import UnavailableGameError
from lutris.gui.dialogs import InputDialog
from lutris.installer import AUTO_ELF_EXE, AUTO_WIN32_EXE
from lutris.installer.installer_file import InstallerFile
from lutris.runners import get_runner_human_name
from lutris.services.base import SERVICE_LOGIN, OnlineService
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util import linux
from lutris.util.downloader import Downloader
from lutris.util.http import HTTPError, Request, UnauthorizedAccessError
from lutris.util.log import logger
from lutris.util.strings import slugify


class ItchIoCover(ServiceMedia):
    """itch.io game cover"""

    service = "itchio"
    size = (315, 250)
    dest_path = os.path.join(settings.CACHE_DIR, "itchio/cover")
    file_patterns = ["%s.png"]

    def get_media_url(self, details: Dict[str, Any]) -> Optional[str]:
        """Extract cover from API"""
        # Animated (gif) covers have an extra field with a png version of the cover
        if "still_cover_url" in details:
            if details["still_cover_url"]:
                return details["still_cover_url"]
        if "cover_url" in details:
            if details["cover_url"]:
                return details["cover_url"]
        else:
            logger.warning("No field 'cover_url' in API game %s", details)
        return None


class ItchIoCoverMedium(ItchIoCover):
    """itch.io game cover, at 60% size"""

    size = (189, 150)


class ItchIoCoverSmall(ItchIoCover):
    """itch.io game cover, at 30% size"""

    size = (95, 75)


class ItchIoGame(ServiceGame):
    """itch.io Game"""

    service = "itchio"

    @classmethod
    def new(cls, igame):
        """Return a Itch.io game instance from the API info"""
        service_game = ItchIoGame()
        service_game.appid = str(igame["id"])
        service_game.slug = slugify(igame["title"])
        service_game.name = igame["title"]
        service_game.details = json.dumps(igame)
        return service_game


class ItchIoService(OnlineService):
    """Service class for itch.io"""

    id = "itchio"
    # According to their branding, "itch.io" is supposed to be all lowercase
    name = _("itch.io")
    icon = "itchio"
    online = True
    drm_free = True
    has_extras = True
    medias = {
        "banner_small": ItchIoCoverSmall,
        "banner_med": ItchIoCoverMedium,
        "banner": ItchIoCover,
    }
    default_format = "banner"

    api_url = "https://api.itch.io"
    login_url = "https://itch.io/login"
    redirect_uris = ["https://itch.io/my-feed", "https://itch.io/dashboard"]
    cache_path = os.path.join(settings.CACHE_DIR, "itchio/api/")

    api_key_path = os.path.join(settings.CACHE_DIR, "itchio/api-key")  # must be kept outside of cache-path
    key_cache_file = os.path.join(cache_path, "profile/owned-keys.json")
    collection_list_cache_file = os.path.join(cache_path, "profile/collections.json")
    games_cache_path = os.path.join(cache_path, "games/")
    collection_cache_path = os.path.join(cache_path, "collections/")
    key_cache = {}

    runners_by_trait = {"p_linux": "linux", "p_windows": "wine"}
    platforms_by_runner = {"wine": "Windows", "linux": "Linux"}

    extra_types = (
        "soundtrack",
        "book",
        "video",
        "documentation",
        "mod",
        "audio_assets",
        "graphical_assets",
        "sourcecode",
        "other",
    )

    def login(self, parent=None):
        # The InputDialog doesn't keep any state we need to protect,
        # but we used to so we'll clear this persistent flag.
        self.is_login_in_progress = False

        question = _(
            "Lutris needs an API key to connect to itch.io. You can obtain one\n"
            "from the itch.io <a href='https://itch.io/user/settings/api-keys'>API keys page</a>.\n\n"
            "You should give Lutris its own API key instead of sharing them."
        )

        api_key_dialog = InputDialog(
            {
                "parent": parent,
                "question": question,
                "title": _("Itch.io API key"),
                "initial_value": "",
            }
        )

        result = api_key_dialog.run()
        if result != Gtk.ResponseType.OK:
            api_key_dialog.destroy()
            self.logout()
            return

        api_key = api_key_dialog.user_value

        if api_key:
            with open(self.api_key_path, "w") as key_file:
                key_file.write(api_key)
            SERVICE_LOGIN.fire(self)
        else:
            self.logout()

    @property
    def credential_files(self):
        """Return a list of all files used for authentication"""
        return [self.api_key_path]

    def load_api_key(self) -> Optional[str]:
        if not os.path.exists(self.api_key_path):
            return None

        with open(self.api_key_path, "r") as key_file:
            return key_file.read()

    def get_headers(self):
        api_key = self.load_api_key()
        if api_key:
            return {"Authorization": f"Bearer {api_key}"}
        else:
            return {}

    def is_connected(self):
        """Check if service is connected and can call the API"""
        if not self.is_authenticated():
            return False
        try:
            profile = self.fetch_profile()
        except (HTTPError, UnauthorizedAccessError):
            logger.warning("Not connected to itch.io account.")
            return False
        return profile and "user" in profile

    def load(self):
        """Load the user's itch.io library"""
        if not self.is_connected():
            logger.error("User not connected to itch.io")
            return

        library = self.get_games()
        games = []
        seen = set()
        for game in library:
            if game["title"] in seen:
                continue
            _game = ItchIoGame.new(game)
            games.append(_game)
            _game.save()
            seen.add(game["title"])
        return games

    def make_api_request(self, path, query=None):
        """Make API request"""
        url = "{}/{}".format(self.api_url, path)
        if query is not None and isinstance(query, dict):
            url += "?{}".format(urlencode(query, quote_via=quote_plus))
        try:
            request = Request(
                url,
                headers=self.get_headers(),
            )
            request.get()
            return request.json
        except UnauthorizedAccessError:
            # We aren't logged in, so we'll log out! This allows you to
            # log in again.
            self.logout()
            raise

    def fetch_profile(self):
        """Do API request to get users online profile"""
        return self.make_api_request("profile")

    def fetch_owned_keys(self, query=None):
        """Do API request to get games owned by user (paginated)"""
        return self.make_api_request("profile/owned-keys", query)

    def fetch_collections(self, query=None):
        """Do API request to users collections"""
        return self.make_api_request("profile/collections", query)

    def fetch_collection(self, collection_id):
        """Do API request to get info about a collection"""
        return self.make_api_request(f"collections/{collection_id}")

    def fetch_collection_games(self, collection_id, query=None):
        """Do API request to get the list of games in a collection"""
        return self.make_api_request(f"collections/{collection_id}/collection-games", query)

    def fetch_game(self, game_id):
        """Do API request to get game info"""
        return self.make_api_request(f"games/{game_id}")

    def fetch_uploads(self, game_id, dl_key):
        """Do API request to get downloadables of a game."""
        query = None
        if dl_key is not None:
            query = {"download_key_id": dl_key}
        return self.make_api_request(f"games/{game_id}/uploads", query)

    def fetch_upload(self, upload, dl_key):
        """Do API request to get downloadable of a game"""
        query = None
        if dl_key is not None:
            query = {"download_key_id": dl_key}
        return self.make_api_request(f"uploads/{upload}", query)

    def fetch_build_patches(self, installed, target, dl_key):
        """Do API request to get game patches"""
        query = None
        if dl_key is not None:
            query = {"download_key_id": dl_key}
        return self.make_api_request(f"builds/{installed}/upgrade-paths/{target}", query)

    def get_download_link(self, upload_id, dl_key):
        """Create download link for installation"""
        url = "{}/{}".format(self.api_url, f"uploads/{upload_id}/download")
        if dl_key is not None:
            query = {"download_key_id": dl_key}
            url += "?{}".format(urlencode(query, quote_via=quote_plus))
        return url

    def get_game_cache(self, appid):
        """Create basic cache key based on game slug and appid"""
        return os.path.join(self.games_cache_path, f"{appid}.json")

    def _cache_games(self, games):
        """Store information about owned keys in cache"""
        os.makedirs(self.games_cache_path, exist_ok=True)
        for game in games:
            filename = self.get_game_cache(game["id"])
            key_path = os.path.join(self.games_cache_path, filename)
            with open(key_path, "w", encoding="utf-8") as cache_file:
                json.dump(game, cache_file)

    def get_owned_games(self, force_load=False):
        """Get all owned library keys from itch.io"""
        owned_keys = []
        fresh_data = True

        if (not force_load) and os.path.exists(self.key_cache_file):
            with open(self.key_cache_file, "r", encoding="utf-8") as key_file:
                owned_keys = json.load(key_file)
            fresh_data = False
        else:
            query = {"page": 1}
            # Basic security; I'm pretty sure itch.io will block us before that tho
            safety = 65507
            while safety:
                response = self.fetch_owned_keys(query)
                if isinstance(response["owned_keys"], list):
                    owned_keys += response["owned_keys"]
                    if len(response["owned_keys"]) == int(response["per_page"]):
                        query["page"] += 1
                    else:
                        break
                else:
                    break
                safety -= 1

            os.makedirs(os.path.join(self.cache_path, "profile/"), exist_ok=True)
            with open(self.key_cache_file, "w", encoding="utf-8") as key_file:
                json.dump(owned_keys, key_file)

        games = []
        for key in owned_keys:
            game = key.get("game", {})
            game["download_key_id"] = key["id"]
            games.append(game)

        if fresh_data:
            self._cache_games(games)
        return games

    def get_collection_cache(self, collection_id):
        """Create basic cache key based on collection slug and collection_id"""
        return os.path.join(self.collection_cache_path, f"{collection_id}.json")

    def _cache_collection(self, collection):
        """Store information about collections in cache"""
        os.makedirs(self.collection_cache_path, exist_ok=True)
        filename = self.get_collection_cache(collection["id"])
        key_path = os.path.join(self.collection_cache_path, filename)
        with open(key_path, "w", encoding="utf-8") as cache_file:
            json.dump(collection, cache_file)

    def get_games_in_collections(self, collection_list: list, force_load=False):
        """Get all games from a list of collections"""

        games = []

        known_appids = set()

        # fetch collected games for each collection
        for collection in collection_list:
            fresh_data = True

            collection_cache_path = self.get_collection_cache(collection["id"])

            if (not force_load) and os.path.exists(collection_cache_path):
                with open(collection_cache_path, "r", encoding="utf-8") as key_file:
                    collection = json.load(key_file)
                fresh_data = False
            else:
                # get the list of games in that collection
                collection["games"] = []
                query = {"page": 1}
                # Basic security; I'm pretty sure itch.io will block us before that tho
                safety = 65507
                while safety:
                    response = self.fetch_collection_games(collection["id"], query)
                    if isinstance(response["collection_games"], list):
                        collection["games"] += response["collection_games"]
                        if len(response["collection_games"]) == int(response["per_page"]):
                            query["page"] += 1
                        else:
                            break
                    else:
                        break
                    safety -= 1

                # filter out bad data for safety
                collection["games"] = [
                    col_game for col_game in collection["games"] if "game" in col_game and "id" in col_game["game"]
                ]

                # try to get download keys from cache
                for col_game in collection["games"]:
                    game = col_game["game"]
                    game_cache_path = self.get_game_cache(game["id"])
                    if (
                        "can_be_bought" in game.get("traits", [])
                        and game.get("min_price", 0) > 0
                        and os.path.exists(game_cache_path)
                    ):
                        with open(game_cache_path, "r", encoding="utf-8") as key_file:
                            cached_game = json.load(key_file)
                            if "download_key_id" in cached_game:
                                game["download_key_id"] = cached_game["download_key_id"]

                # cache the resulting collection
                self._cache_collection(collection)

            if fresh_data:
                self._cache_games([col_game["game"] for col_game in collection["games"]])

            for col_game in collection["games"]:
                game = col_game["game"]
                if game["id"] not in known_appids:
                    known_appids.add(game["id"])
                    games.append(game)
        return games

    def get_collection_list(self, force_load=False):
        collections = []
        if (not force_load) and os.path.exists(self.collection_list_cache_file):
            with open(self.collection_list_cache_file, "r", encoding="utf-8") as key_file:
                collections = json.load(key_file)
        else:
            collections = self.fetch_collections().get("collections", [])
            with open(self.collection_list_cache_file, "w", encoding="utf-8") as key_file:
                json.dump(collections, key_file)
        return collections

    def get_games(self):
        """Return games from the user's library"""
        # get and cache owned games; this populates collections.json for later
        owned_games = self.get_owned_games()

        # get all collections
        collections = self.get_collection_list()

        # if there are collections titled "lutris" (case insestitive) we use only these
        lutris_collections = [col for col in collections if col.get("title", "").casefold() == "lutris" and "id" in col]

        if len(lutris_collections) > 0:
            games = self.get_games_in_collections(lutris_collections)
        else:
            # otherwise, dig up every game we can find!
            games = owned_games + self.get_games_in_collections(collections)

        filtered_games = []
        for game in games:
            classification = game.get("classification")
            if not classification or classification == "game":
                if self._get_detail_runners(game):
                    filtered_games.append(game)
        return filtered_games

    def get_key(self, appid):
        """Retrieve cache information on a key"""
        if not appid:
            raise ValueError("Missing Itch.io app ID")
        game_filename = self.get_game_cache(appid)
        game = {}

        if os.path.exists(game_filename):
            with open(game_filename, "r", encoding="utf-8") as game_file:
                game = json.load(game_file)
        else:
            try:
                game = self.fetch_game(appid).get("game", {})
                self._cache_games([game])
            except HTTPError:
                return

        traits = game.get("traits", [])
        if "can_be_bought" not in traits:
            # If game can not be bought it can not have a key
            return
        if "download_key_id" in game:
            # Return cached key
            return game["download_key_id"]
        if not game.get("min_price", 0):
            # We have no key but the game can be played for free
            return

        # Reload whole key library to check if a key was added
        library = self.get_owned_games(True)
        game = next((x for x in library if x["id"] == appid), game)

        if "download_key_id" in game:
            return game["download_key_id"]
        return

    def get_extras(self, appid):
        """Return a list of bonus content for itch.io game."""
        key = self.get_key(appid)
        try:
            uploads = self.fetch_uploads(appid, key)
        except HTTPError:
            return []
        all_extras = {}
        extras = []
        for upload in uploads["uploads"]:
            if upload["type"] not in self.extra_types:
                continue
            extras.append(
                {
                    "name": upload.get("filename", "").strip().capitalize(),
                    "type": upload.get("type", "").strip(),
                    "total_size": upload.get("size", 0),
                    "id": str(upload["id"]),
                }
            )
        if extras:
            all_extras["Bonus Content"] = extras
        return all_extras

    @staticmethod
    def _get_detail_runners(details: Dict[str, Any], fix_missing_platforms: bool = True) -> List[str]:
        """Extracts the runners available for a given game, given its details.
        This test the traits for specific platforms, and returns the runners
        in a priority order- Linux is first, which occasionally matters.

        Normally, if a game has no platforms we'll assume a default set of runners,
        but 'fix_missing_platforms' may be set to false to turn this off."""
        runners = []
        traits = details.get("traits", [])
        for trait, runner in ItchIoService.runners_by_trait.items():
            if trait in traits:
                runners.append(runner)

        # Special case- some games don't list platform at all. If the game has
        # no "p_" traits- not even "p_osx"- we can assume *all* our platforms are
        # supported and hope for the best!

        if fix_missing_platforms and not runners:
            if not any(t for t in traits if t.startswith("p_")):
                logger.warning(
                    "The itch.io game '%s' has no platforms lists; Lutris will assume all supported runners will work.",
                    details.get("title"),
                )
                return list(ItchIoService.runners_by_trait.values())

        return runners

    def get_installed_slug(self, db_game):
        return db_game["slug"]

    def generate_installer(self, db_game: Dict[str, Any]) -> Dict[str, Any]:
        """Auto generate installer for itch.io game"""
        details = json.loads(db_game["details"])
        runners = self._get_detail_runners(details)

        if runners:
            return self._generate_installer(runners[0], db_game)

        logger.warning("No supported platforms found")
        return {}

    def generate_installers(self, db_game: Dict[str, Any]) -> List[dict]:
        """Auto generate installer for itch.io game"""
        details = json.loads(db_game["details"])

        runners = self._get_detail_runners(details)
        installers = [self._generate_installer(runner, db_game) for runner in runners]

        if len(installers) > 1:
            for installer in installers:
                runner_human_name = get_runner_human_name(installer["runner"])
                installer["version"] += " " + (runner_human_name or installer["runner"])

        return installers

    def _generate_installer(self, runner, db_game: Dict[str, Any]) -> Dict[str, Any]:
        if runner == "linux":
            game_config = {"exe": AUTO_ELF_EXE}
            script = [
                {"extract": {"file": "itchupload", "dst": "$CACHE"}},
                {"merge": {"src": "$CACHE", "dst": "$GAMEDIR"}},
            ]
        elif runner == "wine":
            game_config = {"exe": AUTO_WIN32_EXE}
            script = [{"task": {"name": "create_prefix"}}, {"install_or_extract": "itchupload"}]
        else:
            logger.warning(f"'{runner}' is not a supported runner for itchio")
            return {}

        return {
            "name": db_game["name"],
            "version": "itch.io",
            "slug": db_game["slug"],
            "game_slug": self.get_installed_slug(db_game),
            "runner": runner,
            "itchid": db_game["appid"],
            "script": {
                "files": [{"itchupload": "N/A:Select the installer from itch.io"}],
                "game": game_config,
                "installer": script,
            },
        }

    def get_installed_runner_name(self, db_game: Dict[str, Any]) -> str:
        details = json.loads(db_game["details"])
        runners = self._get_detail_runners(details)
        return runners[0] if runners else ""

    def get_game_platforms(self, db_game: dict) -> List[str]:
        details = json.loads(db_game["details"])

        runners = self._get_detail_runners(details, fix_missing_platforms=False)
        return [self.platforms_by_runner[r] for r in runners]

    def _check_update_with_db(self, db_game, key, upload=None):
        stamp = 0
        if upload:
            uploads = [upload["upload"] if "upload" in upload else upload]
        else:
            uploads = self.fetch_uploads(db_game["service_id"], key)
            if "uploads" in uploads:
                uploads = uploads["uploads"]

        for _upload in uploads:
            # skip extras
            if _upload["type"] in self.extra_types:
                continue
            ts = self._rfc3999_to_timestamp(_upload["updated_at"])
            if (not stamp) or (ts > stamp):
                stamp = ts

        if stamp:
            dbg = games_db.get_games_where(
                installed_at__lessthan=stamp, service=self.id, service_id=db_game["service_id"]
            )
            return len(dbg)
        return False

    def get_update_installers(self, db_game):
        """Check for updates"""
        patch_installers = []
        key = self.get_key(db_game["service_id"])
        upload = None
        outdated = False
        patch_url = None
        info = {}
        info_filename = os.path.join(db_game["directory"], ".lutrisgame.json")
        if os.path.exists(info_filename):
            with open(info_filename, encoding="utf-8") as info_file:
                info = json.load(info_file)
            if "upload" in info:
                # TODO: Implement wharf patching
                # if "build" in info and info["build"]:
                #     upload = self.fetch_upload(info["upload"], key)
                #     patches = self.fetch_build_patches(info["build"], upload["build_id"], key)
                #     patch_urls = []
                #     for build in patches["upgrade_path"]["builds"]:
                #         patch_urls.append("builds/{}/download/patch/default".format(build["id"]))
                # else:
                # Do overinstall of upload / Full build url
                try:
                    upload = self.fetch_upload(info["upload"], key)
                    upload = upload["upload"] if "upload" in upload else upload
                    patch_url = self.get_download_link(info["upload"], key)
                except HTTPError as error:
                    if error.code == 400:
                        # Bad request probably means the upload was removed
                        logger.info("Upload %s for %s seems to be removed.", info["upload"], db_game["name"])
                        outdated = True

                if upload:
                    ts = self._rfc3999_to_timestamp(upload.get("updated_at", 0))
                    if int(info.get("date", 0)) >= ts:
                        return
                    info["date"] = int(datetime.datetime.now().timestamp())

        # Skip time based checks if we already know it's outdated
        if not outdated:
            outdated = self._check_update_with_db(db_game, key, upload)

        if outdated:
            installer = {
                "version": "itch.io",
                "name": db_game["name"],
                "slug": db_game["installer_slug"],
                "game_slug": self.get_installed_slug(db_game),
                "runner": db_game["runner"],
                "script": {
                    "extends": db_game["installer_slug"],
                    "files": [],
                    "installer": [
                        {"extract": {"file": "itchupload", "dst": "$CACHE"}},
                    ],
                },
            }

            if patch_url:
                installer["script"]["files"] = [
                    {
                        "itchupload": {
                            "url": patch_url,
                            "filename": "update.zip",
                            "downloader": lambda f, url=patch_url: Downloader(
                                url, f.download_file, overwrite=True, headers=self.get_headers()
                            ),
                        }
                    }
                ]
            else:
                installer["script"]["files"] = [{"itchupload": "N/A:Select the installer from itch.io"}]

            if db_game["runner"] == "linux":
                installer["script"]["installer"].append(
                    {"merge": {"src": "$CACHE", "dst": "$GAMEDIR"}},
                )
            elif db_game["runner"] == "wine":
                installer["script"]["installer"].append(
                    {"merge": {"src": "$CACHE", "dst": "$GAMEDIR/drive_c/%s" % db_game["slug"]}}
                )

            if patch_url:
                installer["script"]["installer"].append(
                    {"write_json": {"data": info, "file": info_filename, "merge": True}}
                )

            patch_installers.append(installer)
        return patch_installers

    def get_dlc_installers_runner(self, db_game, runner, only_owned=True):
        """itch.io does currently not officially support dlc"""
        return []

    def get_installer_files(self, installer, installer_file_id, selected_extras):
        """Replace the user provided file with download links from itch.io"""

        key = self.get_key(installer.service_appid)
        try:
            uploads = self.fetch_uploads(installer.service_appid, key)
        except HTTPError as ex:
            raise UnavailableGameError from ex
        filtered = []
        extras = []
        files = []
        extra_files = []
        link = None
        filename = "setup.zip"
        selected_extras_ids = set(x["id"] for x in selected_extras or [])

        file = next(_file.copy() for _file in installer.script_files if _file.id == installer_file_id)
        if not file.url.startswith("N/A"):
            link = file.url

        data = {
            "service": self.id,
            "appid": installer.service_appid,
            "slug": installer.game_slug,
            "runner": installer.runner,
            "date": int(datetime.datetime.now().timestamp()),
        }

        if not link or len(selected_extras_ids) > 0:
            for upload in uploads["uploads"]:
                if selected_extras_ids and (upload["type"] in self.extra_types):
                    extras.append(upload)
                    continue
                # default =  games/tools ("executables")
                if upload["type"] == "default" and (installer.runner in ("linux", "wine")):
                    upload_runners = self._get_detail_runners(upload, fix_missing_platforms=False)
                    if installer.runner not in upload_runners:
                        continue

                    is_demo = "demo" in upload["traits"]
                    upload["Weight"] = self.get_file_weight(upload["filename"], is_demo)
                    if upload["Weight"] == 0xFF:
                        continue

                    filtered.append(upload)
                    continue
                # TODO: Implement embedded types: flash, unity, java, html
                # I have not found keys for embdedded games
                # but people COULD write custom installers.
                # So far embedded games can be played directly on itch.io

        if len(filtered) > 0 and not link:
            filtered.sort(key=lambda upload: upload["Weight"])
            # Lutris does not support installer selection
            upload = filtered[0]
            data["upload"] = str(upload["id"])
            if "build_id" in upload:
                data["build"] = str(upload["build_id"])

            link = self.get_download_link(upload["id"], key)
            filename = upload["filename"]

        if link:
            # Adding a file with some basic info for e.g. patching
            installer.script["installer"].append(
                {"write_json": {"data": data, "file": "$GAMEDIR/.lutrisgame.json", "merge": False}}
            )

            files.append(
                InstallerFile(
                    installer.game_slug,
                    installer_file_id,
                    {
                        "url": link,
                        "filename": filename or file.filename or "setup.zip",
                        "downloader": lambda f, url=link: Downloader(
                            url, f.download_file, overwrite=True, headers=self.get_headers()
                        ),
                    },
                )
            )

        for extra in extras:
            if str(extra["id"]) not in selected_extras_ids:
                continue
            link = self.get_download_link(extra["id"], key)
            extra_files.append(
                InstallerFile(
                    installer.game_slug,
                    str(extra["id"]),
                    {
                        "url": link,
                        "filename": extra["filename"],
                        "downloader": lambda f, url=link: Downloader(
                            url, f.download_file, overwrite=True, headers=self.get_headers()
                        ),
                    },
                )
            )

        return files, extra_files

    def get_patch_files(self, installer, installer_file_id):
        """Similar to get_installer_files but for patches"""
        # No really, it is the same! so we just call get_installer_files
        # and strip off the extras files.
        files, _extra_files = self.get_installer_files(installer, installer_file_id, [])
        return files

    def get_file_weight(self, name, demo):
        if name.endswith(".rpm"):
            return 0xFF  # Not supported as an extractor
        # Exclude non-game files that are sometimes miscategorized as "default"
        name_lower = name.lower()
        if any(pattern in name_lower for pattern in ("wallpaper", "background", "artwork", "poster")):
            return 0xFF
        weight = 0x0
        if name.endswith(".deb"):
            weight |= 0x01
        if linux.LINUX_SYSTEM.is_64_bit:
            if "386" in name or "32" in name:
                weight |= 0x08
        else:
            if "64" in name:
                weight |= 0x10
        if demo:
            weight |= 0x40
        return weight

    def get_game_release_date(self, db_game: dict):
        details = db_game.get("details")
        if details:
            details = json.loads(details)
            # Game Release
            release_date = details.get("created_at")
            if release_date is None:
                # Last Update Release
                release_date = details.get("published_at")
            if release_date is not None and isinstance(release_date, str):
                # Return as YYYY-MM-DD
                return release_date[:10]
        return ""

    def _rfc3999_to_timestamp(self, _s):
        # Python does ootb not fully comply with RFC3999; Cut after seconds
        return datetime.datetime.fromisoformat(_s[: _s.rfind(".")]).timestamp()
